import pytest
import ospgrillage as og
import sys, os

sys.path.insert(0, os.path.abspath("../"))


# test if sort vertice function returns a clockwise
def test_sort_vertices():
    point_list = [
        og.LoadPoint(x=8, y=0, z=3, p=5),
        og.LoadPoint(x=8, y=0, z=5, p=5),
        og.LoadPoint(x=5, y=0, z=3, p=5),
        og.LoadPoint(x=5, y=0, z=5, p=5),
    ]
    ref_ans = [
        og.LoadPoint(x=5, y=0, z=3, p=5),
        og.LoadPoint(x=8, y=0, z=3, p=5),
        og.LoadPoint(x=8, y=0, z=5, p=5),
        og.LoadPoint(x=5, y=0, z=5, p=5),
    ]
    actual, _ = og.sort_vertices(point_list)
    assert actual == ref_ans


def test_create_arc_equation():
    a = og.find_circle(
        x1=0,
        y1=0,
        x2=1,
        y2=1,
        x3=2,
        y3=0,
    )

    print(a)

    assert a[0] == [1, 0]
    assert a[1] == 1


def test_rotating_points():
    # check point rotating function
    rotated_coord = og.rotate_point_about_point(
        center_x=0, center_y=-11, angle=32, point=[10, -6]
    )

    print(rotated_coord)
    assert og.np.isclose(rotated_coord[0], 5.585100198856649)
    assert og.np.isclose(rotated_coord[1], -1.3146163850505417)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This test contains a validation against a numerical bridge model made in LUSAS software. Note the test portion of this file are
specific to the comparison i.e. 28m model between LUSAS model outputs and ospg outputs - hence is not advisable to copy-paste
the tests herein to for another new pytest (say another model). However it is reasonable to replicate structure of the wee i.e.
the model creation and running analysis domain of the pytest.
"""
import pytest
import pickle
import json
import numpy as np
import pandas
import sys, os
import ospgrillage as ospg
from pathlib import Path

sys.path.insert(0, os.path.abspath("../"))


def create_json_bridge():
    bridge = {
        "grid": {
            "name": "Super-T 28m",
            "span": 28,
            "width": 10.175,
            "n_longit": 7,
            "n_trans": 11,
            "edge_dist": 1.0875,
            "angle": 0,
        },
        "material": {
            "mat_type": "concrete",
            "code_use": "AS5100-2017",
            "mat_grade": "50MPa",
            "E": 34.8,
            "G": 20,
        },
        "longitudinal": {
            "A": 0.866937,
            "J": 0.154806,
            "Iz": 0.215366,
            "Iy": 0.213602,
            "Az": 0.444795,
            "Ay": 0.258704,
        },
        "edge": {
            "A": 0.044625,
            "J": 0.26253e-3,
            "Iz": 0.241812e-3,
            "Iy": 0.113887e-3,
            "Az": 0.0371929,
            "Ay": 0.0371902,
        },
        "transverse": {
            "A": 0.504,
            "J": 5.22303e-3,
            "Iz": 1.3608e-3,
            "Iy": 0.32928,
            "Az": 0.42,
            "Ay": 0.42,
        },
        "end_slab": {
            "A": 0.252,
            "J": 2.5012e-3,
            "Iz": 0.6804e-3,
            "Iy": 0.04116,
            "Az": 0.21,
            "Ay": 0.21,
        },
        "load": 1,
        "load factors": [1.2, 1.5],
    }
    return bridge


@pytest.fixture()
def create_grillage():
    # Units #
    kilo = 1e3
    milli = 1e-3
    N = 1
    m = 1
    mm = milli * m
    m2 = m**2
    m3 = m**3
    m4 = m**4
    kN = kilo * N
    MPa = N / (mm**2)
    GPa = kilo * MPa

    # read json file for model inputs
    # with open('super_t_28.json') as f:
    #    bridge = json.load(f)

    try:
        with open("super_t_28.json") as f:
            bridge = json.load(f)
    except (FileNotFoundError, IOError):
        bridge = create_json_bridge()
    # material #
    material_prop = bridge["material"]

    # sections (lusas parameters)
    longit_prop = bridge["longitudinal"]
    edge_prop = bridge["edge"]
    trans_prop = bridge["transverse"]
    end_trans_prop = bridge["end_slab"]

    # grid #
    grid_prop = bridge["grid"]

    grid_name = grid_prop["name"]

    L = grid_prop["span"] * m  # span
    w = grid_prop["width"] * m  # width

    n_l = grid_prop["n_longit"]  # number of longitudinal members
    n_t = grid_prop["n_trans"]  # number of transverse members
    edge_dist = (
        grid_prop["edge_dist"] * m
    )  # distance between edge beam and first exterior beam
    angle = grid_prop["angle"]  # skew angle

    # loading #
    P = bridge["load"] * kN

    # load case names (also used as load names) #
    load_name = [
        "Line Test",
        "Points Test (Global)",
        "Points Test (Local)",
        "Patch Test",
        "Moving Two Axle Truck",
    ]

    # load combos (line with patch) #
    load_combo = {
        "name": "Load Combo",
        "factor_1": bridge["load factors"][0],
        "factor_2": bridge["load factors"][1],
    }

    # define material #
    concrete = ospg.create_material(
        material=material_prop["mat_type"],
        code=material_prop["code_use"],
        grade=material_prop["mat_grade"],
    )

    # define sectons #

    longitudinal_section = ospg.create_section(
        A=longit_prop["A"] * m2,
        J=longit_prop["J"] * m3,
        Iz=longit_prop["Iz"] * m4,
        Iy=longit_prop["Iy"] * m4,
        Az=longit_prop["Az"] * m2,
        Ay=longit_prop["Ay"] * m2,
    )

    edge_longitudinal_section = ospg.create_section(
        A=edge_prop["A"] * m2,
        J=edge_prop["J"] * m3,
        Iz=edge_prop["Iz"] * m4,
        Iy=edge_prop["Iy"] * m4,
        Az=edge_prop["Az"] * m2,
        Ay=edge_prop["Ay"] * m2,
    )

    transverse_section = ospg.create_section(
        A=trans_prop["A"] * m2,
        J=trans_prop["J"] * m3,
        Iz=trans_prop["Iz"] * m4,
        Iy=trans_prop["Iy"] * m4,
        Az=trans_prop["Az"] * m2,
        Ay=trans_prop["Ay"] * m2,
        unit_width=True,
    )

    end_tranverse_section = ospg.create_section(
        A=end_trans_prop["A"] * m2,
        J=end_trans_prop["J"] * m3,
        Iz=end_trans_prop["Iz"] * m4,
        Iy=end_trans_prop["Iy"] * m4,
        Az=end_trans_prop["Az"] * m2,
        Ay=end_trans_prop["Ay"] * m2,
    )

    # define grillage members #
    longitudinal_beam = ospg.create_member(
        section=longitudinal_section, material=concrete
    )
    edge_longitudinal_beam = ospg.create_member(
        section=edge_longitudinal_section, material=concrete
    )

    transverse_slab = ospg.create_member(section=transverse_section, material=concrete)
    end_tranverse_slab = ospg.create_member(
        section=end_tranverse_section, material=concrete
    )

    # Create grid #
    simple_grid = ospg.create_grillage(
        bridge_name=grid_name,
        long_dim=L,
        width=w,
        skew=angle,
        num_long_grid=n_l,
        num_trans_grid=n_t,
        edge_beam_dist=edge_dist,
    )

    # assign grillage member to element groups of grillage model #

    simple_grid.set_member(longitudinal_beam, member="interior_main_beam")
    simple_grid.set_member(longitudinal_beam, member="exterior_main_beam_1")
    simple_grid.set_member(longitudinal_beam, member="exterior_main_beam_2")

    simple_grid.set_member(edge_longitudinal_beam, member="edge_beam")

    simple_grid.set_member(transverse_slab, member="transverse_slab")
    simple_grid.set_member(end_tranverse_slab, member="start_edge")
    simple_grid.set_member(end_tranverse_slab, member="end_edge")

    simple_grid.create_osp_model(pyfile=False)
    # ospg.opsplt.plot_model("element")

    return simple_grid, bridge


@pytest.fixture()
def add_analysis_to_simple_grid(create_grillage):
    kilo = 1e3
    milli = 1e-3
    N = 1
    m = 1
    mm = milli * m
    m2 = m**2
    m3 = m**3
    m4 = m**4
    kN = kilo * N
    MPa = N / ((mm) ** 2)
    GPa = kilo * MPa
    simple_grid, bridge = create_grillage

    grid_prop = bridge["grid"]
    L = grid_prop["span"] * m  # span
    w = grid_prop["width"] * m  # width
    n_l = grid_prop["n_longit"]  # number of longitudinal members
    n_t = grid_prop["n_trans"]  # number of transverse members
    edge_dist = (
        grid_prop["edge_dist"] * m
    )  # distance between edge beam and first exterior beam
    angle = grid_prop["angle"]  # skew angle

    ## loading
    P = -bridge["load"] * kN
    # P = - 0.5 * kN

    ## load case names (also used as load names)
    load_name = [
        "Line Test",
        "Points Test (Global)",
        "Points Test (Local)",
        "Patch Test",
        "Moving Two Axle Truck",
    ]

    ## load combos (line with patch)
    load_combo = {
        "name": "Load Combo",
        "factor_1": bridge["load factors"][0],
        "factor_2": bridge["load factors"][1],
    }

    # opsv.plot_model("nodes") # plotting of grid for visualisation

    # Loading #

    # Line load running along midspan width (P is kN/m)
    line_point_1 = ospg.create_load_vertex(x=L / 2, z=0, p=P)
    line_point_2 = ospg.create_load_vertex(x=L / 2, z=w, p=P)
    test_line_load = ospg.create_load(
        loadtype="line", name=load_name[0], point1=line_point_1, point2=line_point_2
    )

    line_case = ospg.create_load_case(name=load_name[0])
    line_case.add_load(test_line_load)
    simple_grid.add_load_case(line_case)

    # Compound Point loads running along midspan at node points

    n_int = n_l - 4  # number of interior members
    space = (w - 2 * edge_dist) / (n_int + 1)  # spacing of interior members

    p_list = [0, edge_dist]
    for s in range(0, n_int):
        p_list.append(edge_dist + space * (s + 1) * m)
    p_list += [w - edge_dist, w]  # include other members nodes

    # in global
    test_points_load = ospg.create_compound_load(name=load_name[1])

    for p in p_list:
        point = ospg.create_load(
            loadtype="point",
            name="Point",
            point1=ospg.create_load_vertex(x=L / 2, z=p, p=P),
        )
        test_points_load.add_load(point)

    points_case = ospg.create_load_case(name=load_name[1])
    points_case.add_load(test_points_load)
    simple_grid.add_load_case(points_case)

    # in local
    test_points_load = ospg.create_compound_load(name=load_name[2])

    for p in p_list:
        point = ospg.create_load(
            loadtype="point",
            name="Point",
            point1=ospg.create_load_vertex(x=0, z=p, p=P),
        )
        test_points_load.add_load(point)

    test_points_load.set_global_coord(ospg.Point(L / 2, 0, 0))
    # shift from local to global

    points_case = ospg.create_load_case(name=load_name[2])
    points_case.add_load(test_points_load)
    simple_grid.add_load_case(points_case)

    # Patch load over entire bridge deck (P is kN/m2)
    patch_point_1 = ospg.create_load_vertex(x=0, z=0, p=P)
    patch_point_2 = ospg.create_load_vertex(x=L, z=0, p=P)
    patch_point_3 = ospg.create_load_vertex(x=L, z=w, p=P)
    patch_point_4 = ospg.create_load_vertex(x=0, z=w, p=P)
    test_patch_load = ospg.create_load(
        loadtype="patch",
        name=load_name[3],
        point1=patch_point_1,
        point2=patch_point_2,
        point3=patch_point_3,
        point4=patch_point_4,
    )

    patch_case = ospg.create_load_case(name=load_name[3])
    patch_case.add_load(test_patch_load)
    simple_grid.add_load_case(patch_case)

    # moving load  2 axle truck (equal loads, 2x2 spacing centre line running)

    # create truck in local coordinate system
    two_axle_truck = ospg.create_compound_load(name=load_name[4])
    point = ospg.create_load(
        loadtype="point",
        name="Point",
        point1=ospg.create_load_vertex(x=0, y=0, z=0, p=P),
    )

    axl_w = 2 * m  # axle width
    axl_s = 2 * m  # axle spacing
    veh_l = axl_s  # vehicle length

    xs = [0, 0, axl_s, axl_s]  # x-axis coordinates
    zs = [0, axl_w, axl_w, 0]  # z-axis coordinates

    for i, _ in enumerate(xs):
        truck_point = ospg.create_load_vertex(x=xs[i], z=zs[i], p=P)
        two_axle_truck.add_load(
            load=ospg.create_load(loadtype="point", name="Point", point1=truck_point)
        )

    # create path object in global coordinate system
    # centre line running of entire span
    # increments so each case is one meter 1 m
    single_path = ospg.create_moving_path(
        start_point=ospg.Point(0 - axl_w, 0, w / 2 - axl_w / 2),
        end_point=ospg.Point(L, 0, w / 2 - axl_w / 2),
        increments=int((L + veh_l + 1) / 1),
    )
    # create moving load (and case)
    moving_truck = ospg.create_moving_load(name=load_name[4])
    # Set path and loads
    moving_truck.set_path(single_path)
    moving_truck.add_load(two_axle_truck)
    simple_grid.add_load_case(moving_truck)  # assign

    ## load combination
    load_combinations = {
        load_name[0]: load_combo["factor_1"],
        load_name[-2]: load_combo["factor_2"],
    }
    simple_grid.add_load_combination(
        load_combination_name=load_combo["name"],
        load_case_and_factor_dict=load_combinations,
    )

    # Run analysis #

    simple_grid.analyze()  # all load cases
    # simple_grid.analyze(load_case=load_name[0]) # specific load case
    # simple_grid.analyze(load_case=load_name[-1]) # specific moving load case

    all_results = simple_grid.get_results()
    # results = simple_grid.get_results(all=True) # same as above
    # results = simple_grid.get_results(load_case=load_case[0]) # specific load case
    move_results = simple_grid.get_results(
        load_case=load_name[-1]
    )  # specific moving load case
    combo_results = simple_grid.get_results(
        combinations=load_combinations
    )  # get combination

    return all_results, move_results, combo_results, bridge


def test_line_load_results(add_analysis_to_simple_grid):
    kilo = 1e3
    milli = 1e-3
    N = 1
    m = 1
    mm = milli * m
    m2 = m**2
    m3 = m**3
    m4 = m**4
    kN = kilo * N
    MPa = N / ((mm) ** 2)
    GPa = kilo * MPa
    # read from json file

    # ospg get results
    all_results, move_results, combo_results, bridge = add_analysis_to_simple_grid
    grid_prop = bridge["grid"]
    L = grid_prop["span"] * m  # span
    w = grid_prop["width"] * m  # width
    n_l = grid_prop["n_longit"]  # number of longitudinal members
    n_t = grid_prop["n_trans"]  # number of transverse members
    edge_dist = (
        grid_prop["edge_dist"] * m
    )  # distance between edge beam and first exterior beam
    angle = grid_prop["angle"]  # skew angle

    ## loading
    P = bridge["load"] * kN

    ## load case names (also used as load names)
    load_name = [
        "Line Test",
        "Points Test (Global)",
        "Points Test (Local)",
        "Patch Test",
        "Moving Two Axle Truck",
    ]

    ## load combos (line with patch)
    load_combo = {
        "name": "Load Combo",
        "factor_1": bridge["load factors"][0],
        "factor_2": bridge["load factors"][1],
    }

    axl_w = 2 * m  # axle width
    axl_s = 2 * m  # axle spacing
    veh_l = axl_s  # vehicle length

    ### Hand calculatioon checks ###

    hand_calcs = (
        [w * P * L / 4]
        + [n_l * P * L / 4] * 2
        + [
            w * P * L**2 / 8,
            2 * P * (L / 2 - axl_s / 2),
            load_combo["factor_1"] * w * P * L / 4
            + load_combo["factor_2"] * w * P * L**2 / 8,
        ]
    )
    # line, point, patch x 3, moving load, combination

    ele_set = list(range(84, 90 + 1))  # midspan members, i is the midspan node
    # static cases
    mid_sta = np.sum(
        np.array(
            all_results["forces"].sel(
                Loadcase=load_name[0:-1], Element=ele_set, Component="Mz_i"
            )
        ),
        axis=1,
    )
    # specific moving load case
    integer = int(L / 2 - 1 + 2)
    mid_mov = np.sum(
        np.array(
            move_results.forces.isel(Loadcase=integer).sel(
                Element=ele_set, Component="Mz_i"
            )
        )
    )

    # load combo
    mid_comb = np.sum(
        np.array(combo_results.forces.sel(Element=ele_set, Component="Mz_i"))
    )

    comp_calcs = list(mid_sta) + [mid_mov] + [mid_comb]

    diff = np.array(hand_calcs) - np.array(
        comp_calcs
    )  # diff between hand calcs and ospg

    # read pickle
    # var = pickle.load(open("Lusas_Outputs_cleaned.p", "rb"))

    # HARD CODE MAPPING - here specific for test_benchmark bridge and its ospg counterpart only
    # create mapping for nodes between LUSAS and ospg
    node_lusas = [
        1,
        3,
        5,
        7,
        9,
        11,
        13,
        15,
        17,
        19,
        21,
        33,
        35,
        37,
        39,
        41,
        43,
        45,
        47,
        49,
        51,
        53,
        65,
        67,
        69,
        71,
        73,
        75,
        77,
        79,
        81,
        83,
        85,
        97,
        99,
        101,
        103,
        105,
        107,
        109,
        111,
        113,
        115,
        117,
        129,
        131,
        133,
        135,
        137,
        139,
        141,
        143,
        145,
        147,
        149,
        161,
        163,
        165,
        167,
        169,
        171,
        173,
        175,
        177,
        179,
        181,
        193,
        195,
        197,
        199,
        201,
        203,
        205,
        207,
        209,
        211,
        213,
    ]
    node_ospg = [
        1,
        15,
        22,
        29,
        36,
        43,
        50,
        57,
        64,
        71,
        8,
        2,
        16,
        23,
        30,
        37,
        44,
        51,
        58,
        65,
        72,
        9,
        3,
        17,
        24,
        31,
        38,
        45,
        52,
        59,
        66,
        73,
        10,
        4,
        18,
        25,
        32,
        39,
        46,
        53,
        60,
        67,
        74,
        11,
        5,
        19,
        26,
        33,
        40,
        47,
        54,
        61,
        68,
        75,
        12,
        6,
        20,
        27,
        34,
        41,
        48,
        55,
        62,
        69,
        76,
        13,
        7,
        21,
        28,
        35,
        42,
        49,
        56,
        63,
        70,
        77,
        14,
    ]
    # create long ele matching dict for long ele , a is LUSAS, b is ospg
    a = [
        127,
        106,
        85,
        64,
        43,
        22,
        1,
        128,
        107,
        86,
        65,
        44,
        23,
        2,
        129,
        108,
        87,
        66,
        45,
        24,
        3,
        130,
        109,
        88,
        67,
        46,
        25,
        4,
        131,
        110,
        89,
        68,
        47,
        26,
        5,
        132,
        111,
        90,
        69,
        48,
        27,
        6,
        133,
        112,
        91,
        70,
        49,
        28,
        7,
        134,
        113,
        92,
        71,
        50,
        29,
        8,
        135,
        114,
        93,
        72,
        51,
        30,
        9,
        136,
        115,
        94,
        73,
        52,
        31,
        10,
    ]
    b = [
        25,
        24,
        23,
        22,
        21,
        20,
        19,
        38,
        37,
        36,
        35,
        34,
        33,
        32,
        51,
        50,
        49,
        48,
        47,
        46,
        45,
        64,
        63,
        62,
        61,
        60,
        59,
        58,
        77,
        76,
        75,
        74,
        73,
        72,
        71,
        90,
        89,
        88,
        87,
        86,
        85,
        84,
        103,
        102,
        101,
        100,
        99,
        98,
        97,
        116,
        115,
        114,
        113,
        112,
        111,
        110,
        129,
        128,
        127,
        126,
        125,
        124,
        123,
        136,
        135,
        134,
        133,
        132,
        131,
        130,
    ]
    match_long_ele = {
        a[i]: b[i] for i in range(len(a))
    }  # dict matching key (lusas ele) to value (ospg ele)

    # extract line load test LUSAS
    # line_load_result = var['3 Line Test Case']
    # extract line load test ospg

    # read from 28m result lusas
    point_lusas_disp_path = [
        "tests",
        "28m results",
        "28m_super_t_displacement",
        "2_Points_Test_Case.csv",
    ]
    line_lusas_disp_path = [
        "tests",
        "28m results",
        "28m_super_t_displacement",
        "3_Line_Test_Case.csv",
    ]
    line_lusas_force_path = [
        "tests",
        "28m results",
        "28m_super_t_forces",
        "3_Line_Test_Case.csv",
    ]
    print(Path.cwd())
    print(Path(__file__))
    point_load_disp_lusas = pandas.read_csv(
        Path(__file__).resolve().parent.parent.joinpath(*point_lusas_disp_path)
    )
    # get ospg point load case
    point_load_disp_ospg = all_results["displacements"].sel(
        Loadcase="Points Test (Global)",
        Component=["x", "y", "z", "theta_x", "theta_y", "theta_z"],
    )

    # lusas elements are 3 noded beam elemnt, this function extracts only the end nodes (first and third) of the model.
    # user to provide node_lusas variable - a list containing the node number correspond to end nodes of beam elements
    lusas_def = reduce_lusas_node_result(
        pd_data=point_load_disp_lusas["DZ[m]"], node_to_extract_list=node_lusas
    )
    sorted_zip_ospg_node = sort_array_by_node_mapping(
        list_of_node=node_ospg,
        data_of_node=point_load_disp_ospg.sel(Component="y").values,
    )

    line_load_disp_lusas = pandas.read_csv(
        Path(__file__).resolve().parent.parent.joinpath(*line_lusas_disp_path)
    )
    lusas_def = reduce_lusas_node_result(
        pd_data=line_load_disp_lusas["DZ[m]"], node_to_extract_list=node_lusas
    )
    line_load_disp_ospg = all_results["displacements"].sel(
        Loadcase="Line Test",
        Component=["x", "y", "z", "theta_x", "theta_y", "theta_z"],
    )
    sorted_zip_ospg_node = sort_array_by_node_mapping(
        list_of_node=node_ospg,
        data_of_node=line_load_disp_ospg.sel(Component="y").values,
    )
    # lusas bending z
    line_load_force_lusas = pandas.read_csv(
        Path(__file__).resolve().parent.parent.joinpath(*line_lusas_force_path)
    )
    single_component_line_lusas = extract_lusas_ele_forces(
        list_of_ele=a, df_force=line_load_force_lusas, component="My[N.m]"
    )
    # ospg bending z
    line_load_bendingz_ospg = (
        all_results["forces"]
        .sel(Loadcase="Line Test", Component=["Mz_i", "Mz_j"], Element=b)
        .values
    )
    # filter only the longitudinal members

    # assert deflection results , if all true/ isclose()
    assert sum(np.isclose(lusas_def, sorted_zip_ospg_node, atol=1e-5)) >= 77
    # assert  # line, point, patch x 3, moving load, combination bending moment about global Z axis close to hand calcs
    assert sum(np.isclose(hand_calcs, np.abs(comp_calcs))) >= 4  # TODO


# ---------------------------------------------
# static methods of test
# function to sort array of nodes based on a provided list of index / numbering. specify if list is either
# index or numbering
def sort_array_by_node_mapping(list_of_node, data_of_node, numbering=True):
    # note: list_of_node.size == data_of_node.size
    if numbering:  # numbering starts from 1
        list_of_node = [d - 1 for d in list_of_node]
        zip_element = zip(list_of_node, data_of_node)
    else:
        zip_element = zip(list_of_node, data_of_node)
    sorted_zip_element = sorted(zip_element)

    return [data_of_node[i] for i in list_of_node]


# function to extract data at node points corresponding to end nodes of element - this function is for model with
# beam elements having 3 or more nodes.
def reduce_lusas_node_result(pd_data, node_to_extract_list):
    return [
        element
        for counter, element in enumerate(pd_data)
        if counter + 1 in node_to_extract_list
    ]


def extract_lusas_ele_forces(list_of_ele, df_force, component: str):
    output = []
    for ele_num in list_of_ele:
        ele_output = []
        for index, row in df_force.iterrows():
            if row.Element == ele_num:
                # check if its intermediate node, if true, continue
                if all([index != 0, index != len(df_force) - 1]):
                    if (
                        not df_force.loc[index - 1].Element
                        == df_force.loc[index + 1].Element
                    ):
                        ele_output.append(row[component])
        output.append(ele_output)

    return output



from fixtures import *

sys.path.insert(0, os.path.abspath("../"))


def test_material_command(ref_bridge_properties):
    # test created material in creating model
    I_beam, slab, exterior_I_beam, concrete = ref_bridge_properties
    # construct grillage model
    example_bridge = og.OspGrillage(
        bridge_name="SuperT_10m",
        long_dim=10,
        width=7,
        skew=-42,
        num_long_grid=7,
        num_trans_grid=5,
        edge_beam_dist=1,
        mesh_type="Ortho",
    )

    # set grillage member to element groups of grillage model
    example_bridge.set_member(I_beam, member="interior_main_beam")
    example_bridge.set_member(exterior_I_beam, member="exterior_main_beam_1")
    example_bridge.set_member(exterior_I_beam, member="exterior_main_beam_2")
    example_bridge.set_member(exterior_I_beam, member="edge_beam")
    example_bridge.set_member(slab, member="transverse_slab")
    example_bridge.set_member(exterior_I_beam, member="start_edge")
    example_bridge.set_member(exterior_I_beam, member="end_edge")

    example_bridge.create_osp_model(pyfile=False)


def test_create_material():
    # test basic functionality of material module
    E = 1000
    G = 100
    v = 0.3
    rho = 1800

    # specific Concrete01 properties
    fpc = 10
    epsc0 = 10
    fpcu = 10
    epsU = 10

    concrete_code = og.create_material(
        material="concrete", code="AS5100-2017", grade="50MPa"
    )
    assert concrete_code.density == 2.4e3

    concrete_custom = og.create_material(E=E, G=G, v=v, rho=rho)  # with complete inputs

    concrete_custom_wo_g = og.create_material(E=E, v=v, rho=rho)  # with missing G input
    expect_g = 384.615
    assert og.np.isclose(concrete_custom_wo_g.shear_modulus, expect_g, rtol=0.1)

    concrete_custom_wo_rho = og.create_material(
        E=E, rho=rho
    )  # with missing G and rho input
    expect_v = 0.3
    assert og.np.isclose(concrete_custom_wo_g.poisson_ratio, expect_v, rtol=0.1)
    # concrete_ops = og.create_material(fpc=fpc,epsc0=epsc0,fpcu=fpcu,epsU=epsU)


# ─────────────────────────────────────────────────────────────────────────────
# Material.get_material_args() – Concrete01 and Steel01 code paths
# ─────────────────────────────────────────────────────────────────────────────


def test_get_material_args_concrete():
    """get_material_args for Concrete01 must return a non-empty arg list."""
    # Codified lookup sets fpc; supply the remaining uniaxialMaterial params
    # (epsc0, fpcu, epsU) that are model-dependent and not stored in the library.
    mat = og.create_material(material="concrete", code="AS5100-2017", grade="50MPa")
    mat.epsc0 = -0.002  # strain at peak compressive stress
    mat.fpcu = -5e6  # crushing stress
    mat.epsU = -0.005  # ultimate strain
    ops_type, args = mat.get_material_args()
    assert ops_type == "Concrete01"
    assert len(args) == 4
    assert None not in args


def test_get_material_args_steel():
    """get_material_args for Steel01 must return a non-empty arg list."""
    # Bypass codified lookup (which has a pre-existing 'fc' KeyError for steel)
    # and build a fully-specified material directly.
    from ospgrillage.material import Material

    mat = Material(E=200e9, v=0.3)
    mat.material_type = "steel"
    mat.ops_mat_type = "Steel01"
    mat.Fy = 500e6
    mat.E0 = 200e9
    mat.b = 0.01
    mat.a1 = 0.04
    mat.a2 = 0.04
    mat.a3 = 0.04
    mat.a4 = 0.04
    ops_type, args = mat.get_material_args()
    assert ops_type == "Steel01"
    assert len(args) == 7
    assert None not in args



# -*- coding: utf-8 -*-
"""
Load module consist of class/methods which wraps ```OpenSeesPy``` to create and run load analysis. First part of the module
comprise the user interface functions. Following are the classes of various load types, compound load, load case,
moving load and moving load path.
"""

import numpy as np
import pprint
from collections import namedtuple
from collections.abc import Iterable
from copy import deepcopy
from typing import Union

from scipy import interpolate

from ospgrillage.mesh import Point, Mesh, create_point
from ospgrillage.utils import (
    find_circle,
    get_distance,
    get_slope,
    get_y_intcp,
    inv_line_func,
    line_func,
    sort_vertices,
)

__all__ = [
    "LoadCase",
    "LoadModel",
    "Loads",
    "MovingLoad",
    "NodalLoad",
    "NodeForces",
    "PatchLoading",
    "Path",
    "PointLoad",
    "LineLoading",
    "LoadVertex",
    "LoadPoint",  # deprecated alias for LoadVertex
    "CompoundLoad",
    "Line",
    "ShapeFunction",
    "create_load_vertex",
    "create_load",
    "create_load_case",
    "create_load_model",
    "create_moving_load",
    "create_moving_path",
    "create_compound_load",
    "create_point",  # re-exported for convenience
]


def create_load_vertex(**kwargs):
    """
    Create a :class:`LoadVertex` — a spatial coordinate with load magnitude.

    Load vertices are the building blocks of every load type.  A point load
    needs one vertex, a line load two, and a patch load four (or more).

    :param x: x coordinate of the load vertex.
    :type x: float or int
    :param y: y coordinate. Defaults to model plane ``y = 0``.
    :type y: float or int
    :param z: z coordinate of the load vertex.
    :type z: float or int
    :param p: Magnitude of vertical load in the y direction.
    :type p: float or int
    :returns: :class:`LoadVertex` ``(x, y, z, p)`` namedtuple.
    :raises ValueError: If one or more of ``x``, ``z``, or ``p`` are missing.
    """
    x = kwargs.get("x", None)
    y = kwargs.get("y", 0)
    z = kwargs.get("z", None)
    p = kwargs.get("p", None)

    if not any([x is None, y is None, z is None, p is None]):
        return LoadVertex(x, y, z, p)
    else:
        raise ValueError("Missing one or more keyword arguments for x=, y=, z=, or p=")


def create_point(**kwargs):
    """
    User interface function to create a Point namedTuple.

    :param x: x coordinate.
    :type x: float or int
    :param y: y coordinate. Defaults to ``0``.
    :type y: float or int
    :param z: z coordinate.
    :type z: float or int
    :returns: ``Point(x, y, z)`` namedTuple.
    """
    x = kwargs.get("x", None)
    y = kwargs.get("y", 0)
    z = kwargs.get("z", None)

    if not any(
        [
            x is None,
            z is None,
        ]
    ):
        return Point(x, y, z)
    else:
        raise ValueError("Missing one or more keyword arguments for x=, z=, ")


def create_load_case(**kwargs):
    """
    Create a load case container for loads to apply in a grillage analysis.

    Use :func:`~ospgrillage.load.LoadCase.add_load` to populate the load case.

    :param name: Name string for the load case.
    :type name: str
    :returns: :class:`~ospgrillage.load.LoadCase` object.
    """
    return LoadCase(**kwargs)


def create_compound_load(**kwargs):
    """
    Create a compound load group that combines multiple individual loads.

    Use :func:`~ospgrillage.load.CompoundLoad.add_load` to add loads to the group.

    :returns: :class:`~ospgrillage.load.CompoundLoad`
    """
    return CompoundLoad(**kwargs)


def create_load(**kwargs):
    """
    User interface function to create load types.

    :param loadtype: Type of load to create. One of ``"point"``, ``"line"``, ``"patch"``, or ``"nodal"``.
    :type loadtype: str
    :param point1: First load vertex in global coordinates (required for all load types).
    :type point1: LoadVertex
    :param point2: Second load vertex in global coordinates (required for line and patch loads).
    :type point2: LoadVertex, optional
    :param point3: Third load vertex (patch loads only).
    :type point3: LoadVertex, optional
    :param point4: Fourth load vertex (patch loads only).
    :type point4: LoadVertex, optional
    :returns: :class:`~ospgrillage.load.PointLoad`, :class:`~ospgrillage.load.LineLoading`,
        :class:`~ospgrillage.load.PatchLoading`, or :class:`~ospgrillage.load.NodalForces`
        depending on ``loadtype``.
    """
    type = kwargs.get("loadtype", None)

    vertice_list = []
    for i in range(4):
        vertice_string = f"point{i + 1}"
        vertice_list.append(kwargs.get(vertice_string, None))

    vertice_count = sum([1 for vertex in vertice_list if vertex])

    if vertice_count == 1:
        return PointLoad(**kwargs)
    elif vertice_count == 2:
        return LineLoading(**kwargs)
    elif vertice_count >= 4:
        return PatchLoading(**kwargs)
    elif type == "nodal":
        fx = kwargs.get("Fx", 0)
        fy = kwargs.get("Fy", 0)
        fz = kwargs.get("Fz", 0)
        mx = kwargs.get("Mx", 0)
        my = kwargs.get("My", 0)
        mz = kwargs.get("Mz", 0)
        tag = kwargs.get("node_tag", None)
        name = kwargs.get("name", None)
        if any(
            [fx is None, fy is None, fz is None, mx is None, my is None, mz is None]
        ):
            raise ValueError(
                "Missing arguments for nodal force definition : Hint check if all required keywords are given"
            )
        force = NodeForces(fx, fy, fz, mx, my, mz)
        return NodalLoad(name=name, node_tag=tag, node_force=force)
    else:
        raise TypeError(
            'load_type must either be "nodal" or number of load points is incorrect. hint:'
            " number of load points must either be 1 (point), 2 (line) or 4 (patch)"
        )


def create_moving_load(**kwargs):
    """
    User interface function to create Moving Load object. Following this function, users are required to:

    #. Set a common path via :func:`~ospgrillage.load.MovingLoad.set_path`.
    #. Add loads via :func:`~ospgrillage.load.MovingLoad.add_load`.

    :param common_path: Path object that all load groups will traverse.
    :type common_path: Path
    :param global_increment: Number of increments for discretizing the Path object.
        Used in advanced cases where a MovingLoad has multiple load groups each with a unique path.
    :type global_increment: float or int, optional
    :returns: :class:`~ospgrillage.load.MovingLoad` object.

    """
    return MovingLoad(**kwargs)


def create_moving_path(**kwargs):
    """
    User interface function to create Path object for moving load.

    :param start_point: Starting point of the path.
    :type start_point: Point
    :param end_point: End point of the path.
    :type end_point: Point
    :param increments: Number of incremental steps along the path. Defaults to ``50``.
    :type increments: int
    :param mid_point: Optional intermediate point to define a curved path. Defaults to ``None``.
    :type mid_point: Point, optional
    :returns: :class:`~ospgrillage.load.Path` object.
    """
    return Path(**kwargs)


# named tuple definitions
LoadVertex = namedtuple("LoadVertex", ["x", "y", "z", "p"])
"""A spatial coordinate with load magnitude — the building block of all loads.

``LoadVertex(x, y, z, p)`` where *x*, *y*, *z* are global coordinates and
*p* is the load magnitude (force, force/length, or force/area depending on
the load type).

.. versionadded:: 0.5.0
"""

#: Deprecated alias for :class:`LoadVertex`.  Prefer ``LoadVertex`` in new code.
LoadPoint = LoadVertex

NodeForces = namedtuple("node_forces", ["Fx", "Fy", "Fz", "Mx", "My", "Mz"])
Line = namedtuple("line", ["m", "c", "phi"])


# ----------------------------------------------------------------------------------------------------------------
# Loading classes
# ---------------------------------------------------------------------------------------------------------------
class Loads:
    """
    Base class for Point, Line, and Patch loads.

    This class provides the common interface and functionality for all load types
    applied to grillage structures. It manages load point definitions (up to 8 points),
    shape function options, and provides methods for load transformations such as
    movement and factor application.

    Attributes
    ----------
    name : str
        Name identifier for the load
    load_point_1 to load_point_8 : LoadVertex
        LoadVertex namedtuples defining spatial coordinates and force magnitudes
    shape_function : str, default 'linear'
        Shape function type used for load interpolation
    point_list : list
        List of all defined load points in order
    compound_dist_x : float
        X-distance offset for compound load grouping (local coordinates)
    compound_dist_z : float
        Z-distance offset for compound load grouping (local coordinates)
    ref_point : Point or None
        Reference point for compound load positioning (local coordinates)
    compound_group : int or None
        Group number for accessing compound load groups in LoadCase
    spec : dict
        Load specification dictionary containing points and metadata
    load_counter : int
        Counter for tracking compound load instances
    """

    load_point_1: LoadVertex
    load_point_2: LoadVertex
    load_point_3: LoadVertex
    load_point_4: LoadVertex
    load_point_5: LoadVertex
    load_point_6: LoadVertex
    load_point_7: LoadVertex
    load_point_8: LoadVertex

    def __init__(self, **kwargs):
        """

        :param name: Name of load
        :param Fx: Axis force in x axis
        :param Fy: Axis force in y axis
        :param Fz: Axis force in z axis
        :param Mx: Moment about x axis
        :param My: Moment about y axis
        :param Mz: Moment about z axis
        :param point1: First load vertex in global coordinates.
        :type point1: LoadVertex, optional
        :param point2: Second load vertex in global coordinates.
        :type point2: LoadVertex, optional
        :param point3: Third load vertex in global coordinates.
        :type point3: LoadVertex, optional
        :param point4: Fourth load vertex in global coordinates (required for patch loads).
        :type point4: LoadVertex, optional
        :param localpoint1: First load vertex in local coordinates.
        :type localpoint1: LoadVertex, optional
        :param localpoint2: Second load vertex in local coordinates.
        :type localpoint2: LoadVertex, optional


        """
        #
        self.name = kwargs.get("name", None)

        # Initialise dict for key load points of line UDL and patch load definitions
        self.load_point_data = dict()
        # parse namedtuple of global coordinates
        self.load_point_1 = kwargs.get("point1", None)
        self.load_point_2 = kwargs.get("point2", None)
        self.load_point_3 = kwargs.get("point3", None)
        self.load_point_4 = kwargs.get("point4", None)
        self.load_point_5 = kwargs.get("point5", None)
        self.load_point_6 = kwargs.get("point6", None)
        self.load_point_7 = kwargs.get("point7", None)
        self.load_point_8 = kwargs.get("point8", None)

        # shape function
        self.shape_function = kwargs.get("shape_function", "linear")
        # check if user skipped point 1 and defined point1 as point 2 instead
        if all([self.load_point_1 is None, self.load_point_2 is not None]):
            raise ValueError("Load point 1 is not defined")

        # list of load points tuple
        self.point_list = [
            self.load_point_1,
            self.load_point_2,
            self.load_point_3,
            self.load_point_4,
            self.load_point_5,
            self.load_point_6,
            self.load_point_7,
            self.load_point_8,
        ]
        # self.local_point_list = [self.local_load_point_1, self.local_load_point_2, self.local_load_point_3,
        #                          self.local_load_point_4, self.local_load_point_5, self.local_load_point_6,
        #                          self.local_load_point_7, self.local_load_point_8]
        # var for compound load group (handled by LoadCase class when creating compound groups)
        self.compound_dist_x = 0  # local coordinate system
        self.compound_dist_z = 0  # local coordinate system
        self.ref_point = None  # local coordinate system
        self.compound_group = None  # group number access by LoadCase class to move load group if any path is defined
        # spec dict
        self.spec = dict(
            name=self.name, global_points=self.point_list, ref_point=self.ref_point
        )  # dict {node number: {Fx:val, Fy:val, Fz:val, Mx:val, My:val, Mz:val}}
        self.load_counter = 0  # counter for compound load

    # function called by Moving load module to move the load group
    def move_load(self, ref_point: Point):
        """
        Function to move each load point of load type by a reference coordinate. This function is handled by OpsGrillage
        :param ref_point: coordinate to be moved
        :type ref_point: namedTuple Point(x,y,z)
        :returns: None. Modifies load points in-place by translating each by ``(+x, +y, +z)`` from ``ref_point``.
        """
        if any(self.point_list):
            self.load_point_1 = (
                self.load_point_1._replace(
                    x=self.load_point_1.x + ref_point.x,
                    z=self.load_point_1.z + ref_point.z,
                )
                if self.load_point_1 is not None
                else self.load_point_1
            )
            self.load_point_2 = (
                self.load_point_2._replace(
                    x=self.load_point_2.x + ref_point.x,
                    z=self.load_point_2.z + ref_point.z,
                )
                if self.load_point_2 is not None
                else self.load_point_2
            )
            self.load_point_3 = (
                self.load_point_3._replace(
                    x=self.load_point_3.x + ref_point.x,
                    z=self.load_point_3.z + ref_point.z,
                )
                if self.load_point_3 is not None
                else self.load_point_3
            )
            self.load_point_4 = (
                self.load_point_4._replace(
                    x=self.load_point_4.x + ref_point.x,
                    z=self.load_point_4.z + ref_point.z,
                )
                if self.load_point_4 is not None
                else self.load_point_4
            )
            self.load_point_5 = (
                self.load_point_5._replace(
                    x=self.load_point_5.x + ref_point.x,
                    z=self.load_point_5.z + ref_point.z,
                )
                if self.load_point_5 is not None
                else self.load_point_5
            )
            self.load_point_6 = (
                self.load_point_6._replace(
                    x=self.load_point_6.x + ref_point.x,
                    z=self.load_point_6.z + ref_point.z,
                )
                if self.load_point_6 is not None
                else self.load_point_6
            )
            self.load_point_7 = (
                self.load_point_7._replace(
                    x=self.load_point_7.x + ref_point.x,
                    z=self.load_point_7.z + ref_point.z,
                )
                if self.load_point_7 is not None
                else self.load_point_7
            )
            self.load_point_8 = (
                self.load_point_8._replace(
                    x=self.load_point_8.x + ref_point.x,
                    z=self.load_point_8.z + ref_point.z,
                )
                if self.load_point_8 is not None
                else self.load_point_8
            )

    def apply_load_factor(self, factor=1):
        """
        Apply a load factor to all load point magnitudes.

        Multiplies the vertical force component (p value) of each defined load point
        by the given factor. Commonly used for load combination analyses and sensitivity
        studies.

        Parameters
        ----------
        factor : float, default 1
            Multiplicative factor to apply to all load point magnitudes

        Returns
        -------
        None
            Modifies load points in-place
        """
        self.load_point_1 = (
            self.load_point_1._replace(p=factor * self.load_point_1.p)
            if self.load_point_1 is not None
            else self.load_point_1
        )
        self.load_point_2 = (
            self.load_point_2._replace(p=factor * self.load_point_2.p)
            if self.load_point_2 is not None
            else self.load_point_2
        )
        self.load_point_3 = (
            self.load_point_3._replace(p=factor * self.load_point_3.p)
            if self.load_point_3 is not None
            else self.load_point_3
        )
        self.load_point_4 = (
            self.load_point_4._replace(p=factor * self.load_point_4.p)
            if self.load_point_4 is not None
            else self.load_point_4
        )
        self.load_point_5 = (
            self.load_point_5._replace(p=factor * self.load_point_5.p)
            if self.load_point_5 is not None
            else self.load_point_5
        )
        self.load_point_6 = (
            self.load_point_6._replace(p=factor * self.load_point_6.p)
            if self.load_point_6 is not None
            else self.load_point_6
        )
        self.load_point_7 = (
            self.load_point_7._replace(p=factor * self.load_point_7.p)
            if self.load_point_7 is not None
            else self.load_point_7
        )
        self.load_point_8 = (
            self.load_point_8._replace(p=factor * self.load_point_8.p)
            if self.load_point_8 is not None
            else self.load_point_8
        )

    def get_magnitude(self):
        """
        Return the force magnitudes of all defined load points.

        Extracts the vertical force component (p value) from each load point
        defined in the load object.

        Returns
        -------
        list
            List of force magnitudes in order of load_point_1 through load_point_8,
            with None values where load points are undefined
        """
        magnitude = []
        for load_point in self.point_list:
            if load_point is not None:
                magnitude.append(load_point.p)

        return magnitude

    def __str__(self):
        return "Load object {} \n".format(self.name) + pprint.pformat(self.spec)


class NodalLoad(Loads):
    """
    Class for Nodal loads.
    """

    def __init__(self, node_tag: int, node_force, name=None):
        """
        Inits the Nodal load class. NodalLoad requires a NodalForce(Fx,Fy,Fz,Mx,My,Mz) as input.

        :param name: Name of load
        :type name: str
        :param node_tag: Node tag of grillage model for nodal load to be applied
        :type node_tag: int
        :param node_force: Named tuple of node forces
        :type node_force: NodalForces(Fx,Fy,Fz,Mx,My,Mz)
        """
        super().__init__(name=name, node_tag=node_tag)
        self.Fx = node_force.Fx
        self.Fy = node_force.Fy
        self.Fz = node_force.Fz
        self.Mx = node_force.Mx
        self.My = node_force.My
        self.Mz = node_force.Mz
        self.node_tag = node_tag
        if not isinstance(node_tag, Iterable):
            node_list = [node_tag]
        else:
            node_list = node_tag
        for nodes in node_list:
            self.spec[nodes] = {
                "Fx": self.Fx,
                "Fy": self.Fy,
                "Fz": self.Fz,
                "Mx": self.Mx,
                "My": self.My,
                "Mz": self.Mz,
            }

    def get_nodal_load_call(self):
        """
        Returns a ``(func_name, args, kwargs)`` tuple for ``ops.load()``.

        This is the preferred form; it lets the caller dispatch via
        ``_OpsProxy._dispatch()`` without any string building or ``eval``.
        """
        load_value = [self.Fx, self.Fy, self.Fz, self.Mx, self.My, self.Mz]
        return ("load", (self.node_tag, *load_value), {})

    def get_nodal_load_str(self):
        """
        Returns an ops.load() command string for the NodalLoad.

        .. deprecated::
            Use :meth:`get_nodal_load_call` instead.  This method is retained
            for backward compatibility only.
        """
        load_value = [self.Fx, self.Fy, self.Fz, self.Mx, self.My, self.Mz]
        load_str = "ops.load({pt}, *{val})\n".format(pt=self.node_tag, val=load_value)
        return load_str


class PointLoad(Loads):
    """
    Class for concentrated point loads.

    Represents a single concentrated force applied at a point on the grillage
    structure. Inherits all attributes and methods from the Loads base class.
    """

    def __init__(self, **kwargs):
        """
        Initialize a PointLoad instance.

        :param name: Name identifier for the point load.
        :type name: str, optional
        :param point1: Location and magnitude of the concentrated load.
        :type point1: LoadVertex
        :param Fx: Force in the global x direction.
        :type Fx: float, optional
        :param Fy: Force in the global y direction.
        :type Fy: float, optional
        :param Fz: Force in the global z direction.
        :type Fz: float, optional
        :param Mx: Moment about the global x axis.
        :type Mx: float, optional
        :param My: Moment about the global y axis.
        :type My: float, optional
        :param Mz: Moment about the global z axis.
        :type Mz: float, optional
        """
        super().__init__(**kwargs)


class LineLoading(Loads):
    """
    Class for distributed line loads (UDL and non-uniform).

    Represents a load distributed along a line on the grillage. Supports both
    straight lines (defined by two endpoints) and curved lines (defined by three
    points forming a circular arc). Provides interpolation methods for computing
    load magnitudes at arbitrary points along the line.
    """

    def __init__(self, **kwargs):
        """
        Initialize a LineLoading instance.

        :param name: Name identifier for the line load.
        :type name: str, optional
        :param point1: Start point of the line load.
        :type point1: LoadVertex
        :param point2: End point of the line load.
        :type point2: LoadVertex
        :param point3: If provided, defines a curved (circular arc) line load through
            point1, point2, and point3. If ``None``, the load follows a straight line.
        :type point3: LoadVertex, optional
        :param long_beam_element_load: Apply load to longitudinal beam elements. Defaults to ``False``.
        :type long_beam_element_load: bool
        :param trans_beam_element_load: Apply load to transverse beam elements. Defaults to ``False``.
        :type trans_beam_element_load: bool
        """
        super().__init__(**kwargs)

        self.long_beam_ele_load_flag = kwargs.get("long_beam_element_load", False)
        self.trans_beam_ele_load_flag = kwargs.get("trans_beam_element_load", False)

        # if three points are defined, set line as curved circular line with point 2 (x2,y2,z2) in the centre of
        # curve
        if self.load_point_3 is not None:  # curve
            # findCircle assumes model plane is y = 0, ignores y input, y in this case is a 2D view of x z plane
            self.d = find_circle(
                x1=self.load_point_1.x,
                y1=self.load_point_1.z,
                x2=self.load_point_2.x,
                y2=self.load_point_2.z,
                x3=self.load_point_3.x,
                y3=self.load_point_3.z,
            )
            # return a function variable
            self.line_end_point = self.load_point_3
        else:  # straight line with 2 points
            self.m, self.phi = get_slope(
                [self.load_point_1.x, self.load_point_1.y, self.load_point_1.z],
                [self.load_point_2.x, self.load_point_2.y, self.load_point_2.z],
            )
            self.c = get_y_intcp(m=self.m, x=self.load_point_1.x, y=self.load_point_1.z)
            self.angle = (
                np.arctan(self.m) if self.m is not None else np.pi / 2
            )  # in radian
            self.line_end_point = self.load_point_2
            # namedTuple Line
            self.line_equation = Line(self.m, self.c, self.phi)
        # else:
        #     raise ValueError("Invalid load points for line load {}".format(self.name))

    def interpolate_udl_magnitude(self, point_coordinate):
        """
        Interpolate load magnitude at a point along the line load.

        Computes the force magnitude at a specified coordinate by linearly
        interpolating between load_point_1 and load_point_2. Uses parametric
        equations for both straight and curved line segments.

        Parameters
        ----------
        point_coordinate : list
            Coordinate [x, y, z] at which to compute the load magnitude

        Returns
        -------
        float
            Interpolated force magnitude (p value) at the given point coordinate

        Notes
        -----
        For straight lines, interpolation uses the parametric equation:
        p(x) = p1 + (x - x1) / (x2 - x1) * (p2 - p1)
        If the line is vertical (x1 == x2), z-coordinate is used instead.
        For curved lines (point3 defined), interpolation logic is not yet implemented.

        Example
        -------
        Line load from point1 (0, 0, 0, p=100) to point2 (10, 0, 0, p=200).
        Magnitude at x=5 would return 150.
        """
        # input: point_coordinate list of [x,y,z]
        pp = None
        # check if line is straight or curve
        if self.load_point_3 is None:  # straight line
            # x[0],z[0] and p[0] shall be reference point for interpolate
            xp = point_coordinate[0]
            yp = point_coordinate[0]  # not used but generated here
            zp = point_coordinate[0]

            # use parametric equation of line in 3D
            v = [
                self.load_point_2.x - self.load_point_1.x,
                self.load_point_2.p - self.load_point_1.p,
                self.load_point_2.z - self.load_point_1.z,
            ]
            if v[0] == 0 and self.load_point_2.x == self.load_point_1.x:
                pp = (zp - self.load_point_1.z) / v[2] * v[1] + self.load_point_1.p
            else:
                pp = (xp - self.load_point_1.x) / v[0] * v[1] + self.load_point_1.p

        elif self.load_point_3 is not None:  # curve
            # TODO for curved line load
            pass
        return pp

    def get_point_given_distance(self, xbar, point_coordinate):
        """
        Compute a shifted point coordinate given a distance offset perpendicular to the line.

        Translates a reference point along the perpendicular direction to the line load
        by a specified distance. Useful for computing load distribution centroids and
        moment calculations.

        Parameters
        ----------
        xbar : float
            Perpendicular distance offset from the line
        point_coordinate : list
            Reference coordinate [x, y, z] on or near the line load

        Returns
        -------
        list
            Shifted coordinate [x', y', z'] after applying the perpendicular offset.
            The y-component is unchanged (model plane assumption).

        .. note::

            Computation uses the line angle ``self.angle`` to decompose the offset into
            x and z components: ``x' = x - xbar*cos(angle)``, ``z' = z - xbar*sin(angle)``.
        """
        # function to return centroid of line load given reference point coordinate (point2) and xbar calculated based
        # on
        z_dis = xbar * np.sin(self.angle)
        x_dis = xbar * np.cos(self.angle)
        # y dis = 0 due to model plane
        new_point = [
            point_coordinate[0] - x_dis,
            point_coordinate[1],
            point_coordinate[2] - z_dis,
        ]
        return new_point

    def get_line_segment_given_x(self, x):
        """
        Compute the z-coordinate on the line for a given x-coordinate.

        Evaluates the line equation z = mx + c at a specified x-value,
        provided x falls within the bounds of the line segment endpoints.

        Parameters
        ----------
        x : float
            X-coordinate at which to compute the corresponding z-coordinate

        Returns
        -------
        float or None
            The z-coordinate (z = mx + c) on the line at the given x-value.
            Returns None if x is outside the line segment bounds or if the
            line is vertical (m is None).

        Notes
        -----
        Valid only for x values between load_point_1.x and line_end_point.x
        (in either direction).
        For vertical lines (infinite slope), this method returns None.
        """
        if self.line_equation.m is None:  # if vertical line
            pass
        else:
            if (
                self.load_point_1.x <= x <= self.line_end_point.x
                or self.load_point_1.x >= x >= self.line_end_point.x
            ):
                return line_func(self.line_equation.m, self.line_equation.c, x)

    def get_line_segment_given_z(self, z):
        """
        Compute the x-coordinate on the line for a given z-coordinate.

        Evaluates the inverse line equation x = (z - c) / m at a specified z-value,
        provided z falls within the bounds of the line segment endpoints.
        Handles vertical lines (infinite slope) as a special case.

        Parameters
        ----------
        z : float
            Z-coordinate at which to compute the corresponding x-coordinate

        Returns
        -------
        float or None
            The x-coordinate on the line at the given z-value.
            For vertical lines, returns the constant x-value (load_point_1.x).
            Returns None if z is outside the line segment bounds.

        Notes
        -----
        Valid only for z values between load_point_1.z and line_end_point.z
        (in either direction).
        """
        if self.line_equation.m is None:  # if vertical line
            if (
                self.load_point_1.z <= z <= self.line_end_point.z
                or self.load_point_1.z >= z >= self.line_end_point.z
            ):
                return self.load_point_1.x
        else:
            if (
                self.load_point_1.z <= z <= self.line_end_point.z
                or self.load_point_1.z >= z >= self.line_end_point.z
            ):
                return inv_line_func(self.line_equation.m, self.line_equation.c, z)


class PatchLoading(Loads):
    """
    Class for Patch loads.

    By default requires at least 4 load point for patch (quadrilateral). Can take up to 8 load points.
    """

    def __init__(self, **kwargs):
        """
        Initialize a PatchLoading instance.

        Requires exactly 4 load points (``point1``–``point4``) supplied in
        counter-clockwise order to define the patch boundary. Up to 8 points
        may be provided for higher-order patches.

        :param point1: First vertex of the patch boundary (CCW order). Required.
        :type point1: LoadVertex
        :param point2: Second vertex of the patch boundary (CCW order). Required.
        :type point2: LoadVertex
        :param point3: Third vertex of the patch boundary (CCW order). Required.
        :type point3: LoadVertex
        :param point4: Fourth vertex of the patch boundary (CCW order). Required.
        :type point4: LoadVertex
        :param point5: Fifth vertex for 8-point (higher-order) patches.
        :type point5: LoadVertex, optional
        :param point6: Sixth vertex for 8-point patches.
        :type point6: LoadVertex, optional
        :param point7: Seventh vertex for 8-point patches.
        :type point7: LoadVertex, optional
        :param point8: Eighth vertex for 8-point patches.
        :type point8: LoadVertex, optional
        :param name: Name identifier for this load.
        :type name: str, optional
        :raises ValueError: If fewer than 4 vertices are supplied, or if the
            vertices do not form a valid counter-clockwise polygon.
        """
        super().__init__(**kwargs)
        if not all(v is None for v in self.point_list):
            a, _ = sort_vertices(
                [
                    self.load_point_2,
                    self.load_point_3,
                    self.load_point_1,
                    self.load_point_4,
                ]
            )
            match = [
                self.load_point_1,
                self.load_point_2,
                self.load_point_3,
                self.load_point_4,
            ]
        else:
            raise ValueError(
                "vertices missing. hint: patch load must have either 4 or 8 vertices"
            )

        # The CCW sort in sort_vertices() starts from the vertex with the
        # smallest positive angle from the centroid, which is not necessarily
        # load_point_1.  We therefore accept any cyclic rotation of the sorted
        # sequence — the points are valid as long as they form a CCW polygon.
        def _is_cyclic_rotation(seq, candidate):
            """Return True if *candidate* is a cyclic rotation of *seq*."""
            n = len(seq)
            if len(candidate) != n:
                return False
            doubled = seq + seq
            return any(doubled[i : i + n] == candidate for i in range(n))

        if not _is_cyclic_rotation(a, match):
            raise ValueError(
                "vertices of patch load give an invalid patch layout: "
                "make sure the four vertices are supplied in counter-clockwise order"
            )
        # get patch min dimension
        self.patch_min_dim = None  # instantiate
        # procedure to define lines of patch
        self._define_patch_edge_lines()

    def _define_patch_edge_lines(self):
        # create each line
        self.line_1 = LineLoading(point1=self.load_point_1, point2=self.load_point_2)
        self.line_2 = LineLoading(point1=self.load_point_2, point2=self.load_point_3)
        self.line_3 = LineLoading(point1=self.load_point_3, point2=self.load_point_4)

        # if only four point is define , patch load is a four point straight line quadrilateral
        if self.load_point_5 is None:
            # create fourth line
            self.line_4 = LineLoading(
                point1=self.load_point_4, point2=self.load_point_1
            )

            # create equation of plane from four straight lines

            # create interpolate object f

            x = np.array(
                [
                    self.load_point_1.x,
                    self.load_point_2.x,
                    self.load_point_3.x,
                    self.load_point_4.x,
                ]
            )
            z = np.array(
                [
                    self.load_point_1.z,
                    self.load_point_2.z,
                    self.load_point_3.z,
                    self.load_point_4.z,
                ]
            )
            p = np.array(
                [
                    self.load_point_1.p,
                    self.load_point_2.p,
                    self.load_point_3.p,
                    self.load_point_4.p,
                ]
            )

        elif self.load_point_4 is None:
            # update line 3
            self.line_3 = LineLoading(
                point1=self.load_point_3, point2=self.load_point_1
            )

            # TODO create equation of plane from 3 points
            # create interpolate object f
            x = np.array(
                [
                    self.load_point_1.x,
                    self.load_point_2.x,
                    self.load_point_3.x,
                ]
            )
            z = np.array(
                [
                    self.load_point_1.z,
                    self.load_point_2.z,
                    self.load_point_3.z,
                ]
            )
            p = np.array(
                [
                    self.load_point_1.p,
                    self.load_point_2.p,
                    self.load_point_3.p,
                ]
            )

        elif self.load_point_8 is not None:
            # TODO
            # point 1 2 3
            # point 3 4 5
            # point 5 6 7
            # point 7 8 1
            pass

        else:
            raise ValueError(
                "Patch load points for patch load {} not valid".format(self.name)
            )
        # create function to get interpolation of p
        self.patch_mag_interpolate = interpolate.LinearNDInterpolator(
            list(zip(x, z)), p
        )
        mod_list = [ls for ls in self.point_list if ls is not None]
        self.patch_min_dim = min(
            [
                get_distance(p1, p2)
                for (p1, p2) in zip(mod_list, mod_list[1:] + [mod_list[0]])
                if all([p1 is not None, p2 is not None])
            ]
        )


# ---------------------------------------------------------------------------------------------------------------
class CompoundLoad:
    """
    Class for Compound load. Used to group different Load objects into a Compound load group.

    When a Load object is pass as an input, CompoundLoad treats the initial positions (load_points) of the Load classes
    as local coordinates. Then CompoundLoad sets the loads "in-place" of the local coordinate. If class input local_coord
    is given, CompoundLoad replaces the coordinates of the initial load points (retaining the magnitude of load points)

    When set_global_coord() function is called, CompoundLoad sets the input global coordinate as the new centroid of the
    compounded load groups. This is done by shifting each local coordinate load point in all load groups under
    CompoundLoad by x (global) and z (global).

    Here are a few relationships between CompoundLoad and other classes

    * CompoundLoad object can have Loads class and its inheritances (Line, Point, Patch)
    * CompoundLoad handles functions of Load classes (e.g. move_load)

    """

    def __init__(self, name: str):
        """
        Initialize a CompoundLoad instance.

        :param name: Name identifier for this compound load group.
        :type name: str
        """
        self.name = name
        self.compound_load_obj_list = []
        self.local_coord_list = []
        self.centroid = Point(0, 0, 0)  # named tuple Point
        self.global_coord = self.centroid

    def add_load(self, load: Loads):
        """
        Adds a :class:`~ospgrillage.load.Load` to compound load group.
        If a local_coord parameter is given, this new local_coord overwrites the coordinates (either local or global) of the load object.

        .. note::
            If load object is defined using local coordinate and local_coord is None, its default local coord precedes.

        :param load: Load object.
        :type load: PointLoad, LineLoading, PatchLoading
        :param local_coord: Local coordinate of load object
        :type local_coord: Point namedTuple

        """
        # update the load obj to be part of compound load by first
        # shifting all load points relative to centroid of defined load class
        # then shifting centroid and load_points relative to A Local Coordinate system
        load_obj_copy = deepcopy(load)
        # check if input load object is valid (local vs global coordinate) system.
        # If local load + local_coord input, raise ValueError
        # if local_coord is not None and any(load_obj_copy.local_point_list):
        #     raise ValueError("{} was defined in local coordinate space. However a `load_coord=` parameter"
        #                      "exist for this load when creating compound load "
        #                      "Hint: Loads defined in local coordinate space does not "
        #                      "need another "
        #                      "`local_coord=` parameter".format(load_obj_copy.name))
        # if global load + local_coord input, raise Value error
        # if local_coord is not None and not any(load_obj_copy.local_point_list):
        #    raise ValueError("{} was defined in global coordinate. However a local_coord= parameter exist for this load"
        #                     ".Hint: Load defined in global coordinate space does not need `local_coord=` "
        #                     "parameter".format(load_obj_copy.name))
        #    pass
        # inputs for compound load are valid, proceed to set compound load
        # if local coord is passed in,
        # if local_coord is not None:  # else points defined as global already
        #     load_obj_copy.form_compound_load(compound_dist_x=local_coord.x, compound_dist_z=local_coord.z)
        #     # then shift load obj relative to global coord (this is the coord of the model) default is 0,0,0 if not set
        #     # by user
        #     load_obj_copy.move_load(self.global_coord)

        self.compound_load_obj_list.append(
            load_obj_copy
        )  # after update, append to list
        # self.local_coord_list.append(local_coord)

    def set_global_coord(self, global_coord: Point):
        """
        Set global coordinate of the compound load with respect to global coordinate system of grillage model.
        The global coordinate is set to all local load points (i.e. it adds the global coord x y z to each local coord)

        :param global_coord: Value of x y z (global coord) to offset the basic coordinate system
        :type global_coord: Point namedTuple
        """
        # overwrite global coordinate
        if global_coord != self.global_coord:
            self.global_coord = global_coord
            # loop each load type in compound load and shift by global_coord
            append_load_list = []
            # once overwritten, update loadpoints in each load obj of compound load by global_coord
            for loads in self.compound_load_obj_list:
                loads.move_load(
                    global_coord
                )  # shift load objs using Load class method move_load()
                append_load_list.append(
                    loads
                )  # append loads which have been moved to new list and
            self.compound_load_obj_list = (
                append_load_list  # overwrite it to class variable
            )


# ---------------------------------------------------------------------------------------------------------------
class LoadCase:
    """
    Container for a named analysis scenario.

    A LoadCase holds one or more :class:`~ospgrillage.load.Loads` or
    :class:`~ospgrillage.load.CompoundLoad` objects that are applied together
    in a single analysis run. Load factors can be assigned per load object
    via :meth:`add_load`.

    A :class:`~ospgrillage.load.MovingLoad` produces a series of LoadCase
    instances, one per increment, representing the load at each position along
    the path.
    """

    def __init__(self, name: str):
        """
        Initialize a LoadCase instance.

        :param name: Name identifier for this load case (e.g. ``"Dead Load"``).
        :type name: str
        """
        self.name = name
        self.load_groups = []  # list of loads
        self.position = None  # init position list
        self.load_command_list = []  # list of openseespy load commands

    def add_load(self, load: Union[Loads, CompoundLoad], **kwargs):
        """
        Add a load or compound load to this load case.

        :param load: The load to add.
        :type load: Loads or CompoundLoad
        :param global_coord_of_load_obj: Required when the load is defined in
            local coordinates. Sets the origin of the local coordinate system
            onto the global coordinate of the grillage.
        :type global_coord_of_load_obj: Point namedtuple, optional
        :param load_factor: Scale factor applied to this load object within
            the load case. Defaults to ``1``.
        :type load_factor: float, optional
        """
        load_dict = dict()
        load_dict.setdefault("load", deepcopy(load))  # create copy of object instance
        # check if load's load points are local points, if True, check if kwargs global coord is provided
        global_coord_of_load_obj: Point = kwargs.get("global_coord_of_load_obj", None)

        load_factor = kwargs.get("load_factor", 1)
        load_dict.setdefault("factor", load_factor)
        self.load_groups.append(load_dict)

    # function for if load groups are to change its ref position due to movement / traversing loads
    # warning : this function is only to be handled by MovingLoad class
    def move_load_group(self, ref_point: Point):
        """
        Translate all load objects in this load case to a new position.

        Repositions all individual and compound loads by moving them to the specified
        reference point. Used by the MovingLoad class to incrementally move load groups
        along a defined path.

        Parameters
        ----------
        ref_point : Point
            Target position (x, y, z) to move the load group's local reference coordinate.
            Loads are translated to this position.

        Returns
        -------
        None
            Modifies load positions in-place for all load_groups in this LoadCase.

        Notes
        -----
        This method is internal to the MovingLoad analysis workflow and should not
        be called directly by users.
        """
        self.position = ref_point
        for load_dict in self.load_groups:
            load_obj = load_dict.get("load")
            if isinstance(load_obj, CompoundLoad):
                for ind_load_obj in load_obj.compound_load_obj_list:
                    ind_load_obj.move_load(self.position)
            else:
                load_obj.move_load(self.position)


# ---------------------------------------------------------------------------------------------------------------
class MovingLoad:
    """
    Main class of moving load case. MovingLoad class parses and creates multiple loadcase object corresponding to
    traversing the input load groups - be it compound or single. Moving load is able to set various path (defined by
    Path class object) to individual load groups.

    """

    def __init__(self, name, **kwargs):
        """

        :param name: Name string of moving load.
        :type name: str
        :param common_path: Path object specifying the common path for all loads to traverse.
        :type common_path: Path, optional
        :param global_increment: Number of increments for discretizing the Path object.
        :type global_increment: int, optional

        .. note::
            ``global_increment`` is used in advanced moving load analysis where a moving load object
            assigns each load type a corresponding unique Path object (which may differ between load groups).
        """
        self.name = name
        self.load_list = []
        self.load_case_dict_list = []  # Variable to access
        self.moving_load_case = (
            []
        )  # Variable recording all created load case for all load group's respective path
        self.static_load_case = []
        # get kwargs
        self.common_path = kwargs.get("common_path", None)
        self.global_increment = kwargs.get("global_increment", None)  # for advance use
        self.parse = False  # flag for if query option is available
        self.incremental_name = None  # init variable of query method

    def set_path(self, path_obj):
        """
        Function to assign/modify the common path variable with a new Path object.
        All loads added later to Moving load object will traverse the same common path object.

        :param path_obj: Path object to specify common path variable.
        :type path_obj: Path

        """
        if self.global_increment is not None:
            raise ValueError(
                "Option for Basic use - common path defined for all added loads in MovingLoad however"
                "a global increment parameter was defined. Hint: Remove global_increment= on creating"
                "moving load object"
            )
        # else, valid input for setting a basic moving load - proceed setting common path variable
        self.common_path = path_obj

    def add_load(self, load: Union[Loads, CompoundLoad], path_obj=None):
        """
        Set a load type with its path.  Accepts compound load objects,
        which in turn sets the path object to all loads within the compound
        load group.

        :param load: Load or compound-load object.
        :type load: Loads or CompoundLoad
        :param path_obj: Path class object — for advanced use, where users
            specify a unique path object for each load within the moving load.

        """
        # if no path object is added, set empty list to path_obj. The load group will be treated as a static load
        # present throughout the movement of other load groups (added to the series of moving load case)

        load_pair_path = dict()
        load_pair_path.setdefault("load", load)
        # check if basic moving load case
        if self.common_path and self.global_increment is None:
            load_pair_path.setdefault(
                "path", self.common_path.get_path_points()
            )  # .get_path_points() class method of Path class
        elif (
            self.global_increment
        ):  # advance use - where each object has individual path
            load_pair_path.setdefault(
                "path", path_obj.get_custom_path_points(self.global_increment)
            )
        else:  # error, no global statement was provided,
            raise ValueError(
                "No set_path() for moving load {}: Hint run set_path() before add_loads()".format(
                    self.name
                )
            )
        self.load_case_dict_list.append(load_pair_path)

    # function to create incremental load cases for each step of the moving loads. Function handled by OspGrillage
    def parse_moving_load_cases(self):
        """
        Generate individual LoadCase objects for each position along the moving load path.

        Processes all load-path pairs added to this MovingLoad object and creates
        a series of LoadCase instances, one for each increment in the path(s).
        Static loads (with empty paths) are identified and added to all incremental
        load cases. Moving loads are repositioned at each path increment.

        Returns
        -------
        None
            Populates self.moving_load_case list with LoadCase objects corresponding
            to each path increment.

        Notes
        -----
        This method is called internally during the analysis setup and creates all
        the incremental load cases necessary for a moving load analysis. The load cases
        are accessible via self.moving_load_case attribute.
        Static loads (loads without a path) are included in every incremental load case.
        """
        # loop through all load-path pairs and identify static loads
        for load_pair_dict in self.load_case_dict_list:
            if not load_pair_dict["path"]:  # empty path, load is static
                self.static_load_case.append(load_pair_dict["load"])

        # create load case obj for each step in the move
        for load_pair_dict in self.load_case_dict_list:
            path_list = load_pair_dict["path"]  # extract path_list of load object
            load_obj = load_pair_dict["load"]
            # loop to create a load case for each increment of the path obj
            load_case_list = []
            for steps in path_list:
                load_step_lc = LoadCase(
                    name="{} at global position [{:.2f},{:.2f},{:.2f}]".format(
                        self.name, steps[0], steps[1], steps[2]
                    )
                )  # _lc in name stands for load case
                load_obj_copy = deepcopy(
                    load_obj
                )  # Use deepcopy module to copy instance of load
                load_step_lc.add_load(
                    load_obj_copy
                )  # and add load to newly created load case
                # add entries of static load to load groups
                step_point = Point(
                    steps[0], steps[1], steps[2]
                )  # convert increment position into Point tuple
                load_step_lc.move_load_group(
                    step_point
                )  # increment the load groups by step point
                # static load not used
                for (
                    static_load
                ) in (
                    self.static_load_case
                ):  # add static load portions to each incremental load case
                    static_load_copy = deepcopy(static_load)
                    load_step_lc.add_load(static_load_copy)
                load_case_list.append(load_step_lc)
            self.moving_load_case.append(load_case_list)
            self.parse = True
        return self.moving_load_case

    def query(self, incremental_lc_name, **kwargs):
        """
        Function to query properties of moving load

        :param incremental_lc_name: Name string of the incremental load case to query.
        :type incremental_lc_name: str
        :param load_group_name: Name of the load group to query. Defaults to all groups.
        :type load_group_name: list, optional
        :param index: Index of the result entry to return. Defaults to ``0``.
        :type index: int, optional
        :param option: Query option — e.g. ``"position"``. Defaults to ``"position"``.
        :type option: str, optional
        """
        # get query options
        if not self.parse:
            raise ValueError(
                "Moving load is not yet set to a grillage model - no information ready for query. hint"
                "add moving load to a grillage model via add_load_case()"
            )
        # specific incremental load case name
        self.incremental_name = incremental_lc_name
        # specific load group name
        load_group_name = kwargs.get("load_group_name", [])
        index = kwargs.get("index", 0)
        option = kwargs.get("option", "position")

        # instantiate variables
        query_load_case = []
        # get incremental load cases based on incremental names provided by users
        for ml in self.moving_load_case:
            for incremental_lc in ml:
                if self.incremental_name == incremental_lc.name:
                    query_load_case = incremental_lc
                    break
            break
        # get load groups within moving load and its respective path
        if load_group_name:
            selected_load_groups = [
                load
                for load in query_load_case.load_groups
                if load["load"].name in load_group_name
            ]
            selected_lc_list = [
                a for a in self.load_case_dict_list if a["load"].name in load_group_name
            ]
        else:  # select all load groups in
            selected_load_groups = query_load_case.load_groups
            selected_lc_list = self.load_case_dict_list

        # TODO add more options* and check if compound or basic
        if option == "position":
            return [a["load"].point_list for a in selected_load_groups]
        elif option == "offset":
            return [
                a["load"].point_list - b
                for (a, b) in zip(selected_load_groups, selected_lc_list)
            ]
        elif option == "original":
            return selected_lc_list
        elif option == "path":
            return [load["path"] for load in selected_lc_list]


class Path:
    """
    Class for defining and managing moving load paths.

    Represents a trajectory along which loads traverse during a moving load analysis.
    Supports linear paths between start and end points with configurable discretization.
    Provides methods to generate path points with different levels of refinement.
    """

    def __init__(
        self,
        start_point: Point,
        end_point: Point,
        increments: int = 50,
    ):
        """
        Initialize a Path instance.

        Parameters
        ----------
        start_point : Point
            Starting position (x, y, z) of the path
        end_point : Point
            Ending position (x, y, z) of the path
        increments : int, default 50
            Number of discrete steps to divide the path into.
            Higher values give finer resolution for load positioning.

        Notes
        -----
        The path is created as a straight line in 3D space from start_point to end_point.
        Each coordinate is linearly interpolated using numpy.linspace.
        """
        self.start_point = start_point
        self.end_point = end_point
        # here create a straight path
        self.path_points_x = np.linspace(start_point.x, end_point.x, increments)
        self.path_points_y = np.linspace(start_point.y, end_point.y, increments)
        self.path_points_z = np.linspace(start_point.z, end_point.z, increments)
        self.path_points_list = []

    def get_path_points(self) -> list:
        """
        Get the discretized path points for the defined path.

        Generates a list of [x, y, z] coordinates along the straight line path
        from start_point to end_point, using the number of increments specified
        during initialization.

        Returns
        -------
        list of list
            List of path position coordinates [[x0, y0, z0], [x1, y1, z1], ...].
            Length equals the number of increments specified at initialization.
        """
        self.path_points_list = [
            [x, y, z]
            for (x, y, z) in zip(
                self.path_points_x, self.path_points_y, self.path_points_z
            )
        ]
        return self.path_points_list

    def get_custom_path_points(self, new_increment):
        """
        Get path points with a custom number of increments.

        Generates a new set of discretized path points using a different number of
        increments than specified during Path initialization. Useful for refinement
        or coarsening of the path discretization for advanced moving load analyses.

        Parameters
        ----------
        new_increment : int
            New number of increments to discretize the path with.
            Overrides the increments value from initialization.

        Returns
        -------
        list of list
            List of path position coordinates [[x0, y0, z0], [x1, y1, z1], ...].
            Length equals new_increment.

        Notes
        -----
        This method recalculates path points without modifying the Path object's
        internal state. The original increments value remains unchanged.
        """
        path_points_x = np.linspace(self.start_point.x, self.end_point.x, new_increment)
        path_points_y = np.linspace(self.start_point.x, self.end_point.x, new_increment)
        path_points_z = np.linspace(self.start_point.x, self.end_point.x, new_increment)
        path_point_list = [
            [x, y, z] for (x, y, z) in zip(path_points_x, path_points_y, path_points_z)
        ]
        return path_point_list


# ---------------------------------------------------------------------------------------------------------------
def create_load_model(**kwargs):
    """
    Create a load model object representing a vehicle load model.

    :returns: LoadModel object
    """
    return LoadModel(**kwargs)


class LoadModel:
    """
    Class to handle load model generator. This contains library of load models and creates load model using
    `CompoundLoad` class.

    For users wishing to contribute/add load models, do so herein by adding the load model as a function to
    this class.

    """

    def __init__(self, gap=0, **kwargs):
        """
        Init the class.

        :param gap: Gap between axles (specific to M1600 load model). Defaults to ``0``.
        :type gap: float or int
        :param model_type: Load model type identifier.
        :type model_type: str, optional
        :param units: Unit system to use. Defaults to ``"SI"``.
        :type units: str, optional
        :param origin: Origin point for the load model in global coordinates. Defaults to ``Point(0, 0, 0)``.
        :type origin: Point namedTuple, optional
        """
        self.gap = gap
        self.model_type = kwargs.get("model_type", None)
        self.units = kwargs.get("units", "SI")
        self.origin = kwargs.get("origin", Point(0, 0, 0))
        # create unit dict
        self.unit_dict = {"m": 1, "kN": 1e3}
        # create variables
        self.x_offset = self.origin.x
        self.z_offset = self.origin.z
        # default y offset is zero

        # checks
        if self.origin is None:
            self.x_offset = 0

    def create(self):
        """
        Generate a CompoundLoad object for the specified load model type.

        Creates a ready-to-use compound load representing a standard vehicle load model
        with proper axle spacing and weight distribution according to the specified
        standard (e.g., AS5100 M1600).

        Returns
        -------
        CompoundLoad
            A CompoundLoad object representing the vehicle load model with all axles,
            spacing, and load magnitudes configured according to the model_type and
            parameters set during initialization.

        Raises
        ------
        ValueError
            If model_type is not recognized or not implemented.

        Notes
        -----
        Currently supports "M1600" (Australian Standard AS5100).
        Additional load models can be added by implementing new create_* methods.
        """
        if self.model_type == "M1600":
            return self.create_m1600_vehicle(self.gap)

    def create_m1600_vehicle(self, gap):
        """
        AS5100 Australian load model.

        :param gap: Gap between axle group
        :returns: :class:`~ospgrillage.load.CompoundLoad` object of a M1600 vehicle in local coordinate.
        """
        # default SI units
        m = 1
        kN = 1e3
        if self.units == "SI":  # if SI, de-activate converter variables
            ft_convert = 1
            ton_convert = 1
        else:  # Imperial units
            ft_convert = 3.28084
            ton_convert = 0.10036113565668

        axle_dist = 1.25 * m * ft_convert
        left_group_dist = 3.75 * m * ft_convert
        right_group_dist = 5 * m * ft_convert
        wheel_load = 60 * kN * ton_convert
        load_positions_x = [
            0,
            axle_dist,
            axle_dist * 2,
            left_group_dist + axle_dist * 2,
            left_group_dist + axle_dist * 2 + axle_dist,
            left_group_dist + axle_dist * 2 + 2 * axle_dist,
            gap + left_group_dist + axle_dist * 2,
            gap + left_group_dist + axle_dist * 2 + axle_dist,
            gap + left_group_dist + axle_dist * 2 + 2 * axle_dist,
            right_group_dist + gap + left_group_dist + axle_dist * 2,
            right_group_dist + gap + left_group_dist + axle_dist * 2 + axle_dist,
            right_group_dist + gap + left_group_dist + axle_dist * 2 + 2 * axle_dist,
        ]
        load_positions_x = [point + self.x_offset for point in load_positions_x]

        load_positions_z = [-1, 1]
        load_positions_z = [point + self.z_offset for point in load_positions_z]

        M1600_vehicle = create_compound_load(name="M1600 Vehicle")
        for z in load_positions_z:
            for x in load_positions_x:
                vert = create_load_vertex(x=x, z=z, p=wheel_load)
                point = create_load(loadtype="point", name="M1600 point", point1=vert)
                M1600_vehicle.add_load(point)

        return M1600_vehicle


# ---------------------------------------------------------------------------------------------------------------
class ShapeFunction:
    """
    Class for shape functions.

    To add more shape functions:
    * add the string options get_shape_function()
    * defining the shape function as a class function which takes eta and zeta as inputs

    """

    def __init__(
        self, option_three_node: str = "triangle_linear", option_four_node="hermite"
    ):
        """Init the ShapeFunction class. Each shape function can either be triangular or quadrilateral"""
        self.option_three_node = option_three_node
        self.option_four_node = option_four_node

    def get_shape_function(self, option: str, eta: float = 0, zeta: float = 0):
        """
        Get a shape function evaluator for the specified interpolation type.

        Returns a callable that computes shape function values at a given parametric
        location (eta, zeta) within an element. Supports hermite (cubic), linear, and
        triangular linear shape functions.

        Parameters
        ----------
        option : str
            Type of shape function. Options are:
            - 'hermite': 2D cubic Hermite shape functions (4-node quad)
            - 'linear': 2D bilinear shape functions (4-node quad)
            - 'triangle_linear': Linear triangular shape functions (3-node triangle)
        eta : float, default 0
            Parametric coordinate in the first direction (range typically [-1, 1])
        zeta : float, default 0
            Parametric coordinate in the second direction (range typically [-1, 1])

        Returns
        -------
        callable
            Lambda function that evaluates the selected shape function at (eta, zeta).
            Call the returned function with no arguments to get shape function values.

        Notes
        -----
        For hermite and linear options, eta and zeta range from -1 to 1 and represent
        normalized coordinates within the element.
        """
        if option == "hermite":
            return lambda: self.hermite_shape_function_2d(eta, zeta)
        elif option == "linear":
            return lambda: self.linear_shape_function(eta, zeta)
        elif option == "triangle_linear":
            return lambda: self.linear_triangular

    @staticmethod
    def hermite_shape_function_1d(
        zeta: float, a: float
    ):  # using zeta and a as placeholders for normal coor + length of edge element
        # hermite shape functions
        """
        1D hermite shape function

        :param zeta: Normalised position along the element (0 to 1).
        :type zeta: float
        :param a: Element length.
        :type a: float
        :returns: Four Hermite shape function values ``[N1, N2, N3, N4]``.
        """
        N1 = 1 - 3 * zeta**2 + 2 * zeta**3
        N2 = (zeta - 2 * zeta**2 + zeta**3) * a
        N3 = 3 * zeta**2 - 2 * zeta**3
        N4 = (-(zeta**2) + zeta**3) * a
        return [N1, N2, N3, N4]

    @staticmethod
    def hermite_shape_function_2d(eta: float, zeta: float):
        """
        Compute 2D Hermite cubic shape functions and their derivatives.

        Calculates Hermite shape functions for a 4-node quadrilateral element.
        Returns shape function values and derivatives needed for element-level
        load distribution in finite element analysis.

        Parameters
        ----------
        eta : float
            Parametric coordinate in the x-direction (range [-1, 1]).
            eta = -1 at left edge, eta = 1 at right edge.
        zeta : float
            Parametric coordinate in the z-direction (range [-1, 1]).
            zeta = -1 at bottom edge, zeta = 1 at top edge.

        Returns
        -------
        tuple of list
            Three lists:
            - Nv : Vertical shape functions [N1, N2, N3, N4] for displacement/force
            - Nmx : Shape functions [N1, N2, N3, N4] for moment about x-axis (derivative w.r.t. zeta)
            - Nmz : Shape functions [N1, N2, N3, N4] for moment about z-axis (derivative w.r.t. eta)

        .. note::

            Node ordering (counter-clockwise from bottom-left)::

                4 o-----o 3
                  |     |
                  |     |
                1 o-----o 2

            Each shape function Ni equals 1 at node i and 0 at other nodes.
            Hermite functions are cubic polynomials providing C1 continuity across elements.
        """
        # nodes are ordered counter clockwise such that node 1 (n1), is left bottom of relative grid
        # 4 o - - - o 3
        #   |       |
        #   |       |
        # 1 o - - - o 2
        h1 = 0.25 * (2 - 3 * eta + eta**3)
        h2 = 0.25 * (1 - eta - eta**2 + eta**3)
        h3 = 0.25 * (2 + 3 * eta - eta**3)
        h4 = 0.25 * (-1 - eta + eta**2 + eta**3)
        z1 = 0.25 * (2 - 3 * zeta + zeta**3)
        z2 = 0.25 * (1 - zeta - zeta**2 + zeta**3)
        z3 = 0.25 * (2 + 3 * zeta - zeta**3)
        z4 = 0.25 * (-1 - zeta + zeta**2 + zeta**3)
        Nv = [h1 * z1, h3 * z1, h3 * z3, h1 * z3]
        Nmz = [h2 * z1, h4 * z1, h4 * z3, h2 * z3]
        Nmx = [h1 * z2, h3 * z2, h3 * z4, h1 * z4]
        return Nv, Nmx, Nmz

    @staticmethod
    def linear_shape_function(eta, zeta):
        """
        2D linear beam shape function

        :param zeta: Normalised position in the x-direction.
        :type zeta: float
        :param eta: Normalised position in the z-direction.
        :type eta: float
        :returns: Four linear shape function values ``[N1, N2, N3, N4]``.

        .. note::

            Further validation needed — trial on different bridge models.
        """
        N1 = 0.25 * (1 - eta) * (1 - zeta)
        N2 = 0.25 * (1 + eta) * (1 - zeta)
        N3 = 0.25 * (1 + eta) * (1 + zeta)
        N4 = 0.25 * (1 - eta) * (1 + zeta)
        return [N1, N2, N3, N4]

    @staticmethod
    def linear_triangular(x, z, x1, z1, x2, z2, x3, z3):
        """
        Compute 2D linear triangular shape functions.

        Calculates linear (first-order) shape functions for a 3-node triangular element.
        Uses barycentric coordinate formulation to interpolate values within the triangle.

        Parameters
        ----------
        x : float
            X-coordinate of the evaluation point
        z : float
            Z-coordinate of the evaluation point
        x1, z1 : float
            X, Z coordinates of node 1
        x2, z2 : float
            X, Z coordinates of node 2
        x3, z3 : float
            X, Z coordinates of node 3

        Returns
        -------
        list
            Three shape function values [N1, N2, N3] representing the interpolated
            contribution of each node. Sum(N_i) = 1.0 for points inside the triangle.

        .. note::

            The linear triangular element uses the formulation
            ``N_i = (a_i + b_i*x + c_i*z) / (2*A)``
            where ``A`` is the triangle area and ``(a_i, b_i, c_i)`` are computed from node coordinates.
            Points outside the triangle will have negative shape function values.
            The modelling plane is y = constant (XZ plane).
        """
        # modelling plane = y plane
        ae = 0.5 * ((x2 * z3 - x3 * z2) + (z2 - z3) * x1 + (x3 - x2) * z1)
        a1 = (x2 * z3 - x3 * z2) / (2 * ae)
        b1 = (z2 - z3) / (2 * ae)
        c1 = (x3 - x2) / (2 * ae)

        a2 = (x3 * z1 - x1 * z3) / (2 * ae)
        b2 = (z3 - z1) / (2 * ae)
        c2 = (x1 - x3) / (2 * ae)

        a3 = (x1 * z2 - x2 * z1) / (2 * ae)
        b3 = (z1 - z2) / (2 * ae)
        c3 = (x2 - x1) / (2 * ae)
        N1 = a1 + b1 * x + c1 * z
        N2 = a2 + b2 * x + c2 * z
        N3 = a3 + b3 * x + c3 * z
        return [N1, N2, N3]



from ospgrillage import __version__ as version
from fixtures import *

sys.path.insert(0, os.path.abspath("../"))


def test_envelope(bridge_model_42_negative):
    # test functionality of Envelope class and output
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    # create reference line load
    p = 10000
    p2 = 20000
    p3 = 30000  # duplicate of 2nd but with different mag
    barrierpoint_1 = og.create_load_vertex(x=5, z=1, p=p)
    barrierpoint_2 = og.create_load_vertex(x=10, z=7, p=p)
    barrierpoint_3 = og.create_load_vertex(x=10, z=2, p=p2)
    barrierpoint_4 = og.create_load_vertex(x=5, z=5, p=p2)
    barrierpoint_5 = og.create_load_vertex(x=10, z=2, p=p3)
    barrierpoint_6 = og.create_load_vertex(x=5, z=5, p=p3)
    # add moving load case
    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(2, 0, 2, 50)
    )  # Single point load 50 N
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
    )
    Barrier2 = og.create_load(
        loadtype="line", name="Barrieload", point1=barrierpoint_3, point2=barrierpoint_4
    )
    Barrier3 = og.create_load(
        loadtype="line", name="Barrieload", point1=barrierpoint_5, point2=barrierpoint_6
    )
    Patch1 = og.create_load(
        loadtype="patch",
        point1=barrierpoint_1,
        point2=barrierpoint_3,
        point3=barrierpoint_2,
        point4=barrierpoint_4,
    )

    barrierpoint_1 = og.create_load_vertex(x=6, z=2, p=0)
    barrierpoint_2 = og.create_load_vertex(x=11, z=8, p=0)
    barrierpoint_3 = og.create_load_vertex(x=11, z=3, p=0)
    barrierpoint_4 = og.create_load_vertex(x=6, z=6, p=0)

    Patch2 = og.create_load(
        loadtype="patch",
        point1=barrierpoint_1,
        point2=barrierpoint_3,
        point3=barrierpoint_2,
        point4=barrierpoint_4,
    )
    # create basic load case
    barrier_load_case = og.create_load_case(name="Barrier")
    # barrier_load_case.add_load(Barrier)  # ch
    barrier_load_case.add_load(Patch1)  # ch
    # 2nd
    barrier_load_case2 = og.create_load_case(name="Barrier2")
    # barrier_load_case2.add_load(Barrier2)
    barrier_load_case2.add_load(Patch2)
    # 3rd
    # barrier_load_case3 = og.create_load_case(name="Barrier3")
    # barrier_load_case3.add_load(Barrier3)

    # adding load cases to model
    example_bridge.add_load_case(barrier_load_case)
    example_bridge.add_load_case(barrier_load_case2)
    # example_bridge.add_load_case(barrier_load_case3)

    single_path = og.create_moving_path(
        start_point=og.Point(2, 0, 2), end_point=og.Point(4, 0, 3), increments=3
    )  # create path object
    move_point = og.create_moving_load(name="single_moving_point")
    move_point.set_path(single_path)
    move_point.add_load(front_wheel)
    example_bridge.add_load_case(move_point)

    example_bridge.analyze()
    results = example_bridge.get_results()
    comb_results = example_bridge.get_results(
        combinations={"Barrier": 1, "single_moving_point": 2}
    )
    # maxY = results.sel(Component='dy').max()
    envelope = og.create_envelope(
        ds=comb_results, load_effect="dy", array="displacements"
    )
    max_disp = envelope.get()
    print(max_disp)
    move_point.query(
        incremental_lc_name="single_moving_point at global position [2.00,0.00,2.00]"
    )
    # print(comb_results)
    print(og.ops.nodeDisp(25)[1])


def test_plot_force(bridge_model_42_negative):
    # test functionality of plot_force and its output
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative
    # og.opsv.plot_model()
    # og.plt.show()
    # create reference line load
    p = 10000
    p2 = 20000
    p3 = 30000  # duplicate of 2nd but with different mag
    barrierpoint_1 = og.create_load_vertex(x=5, z=1, p=p)
    barrierpoint_2 = og.create_load_vertex(x=10, z=7, p=p)
    barrierpoint_3 = og.create_load_vertex(x=10, z=2, p=p2)
    barrierpoint_4 = og.create_load_vertex(x=5, z=5, p=p2)
    barrierpoint_5 = og.create_load_vertex(x=10, z=2, p=p3)
    barrierpoint_6 = og.create_load_vertex(x=5, z=5, p=p3)
    # add moving load case
    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(2, 0, 2, 50)
    )  # Single point load 50 N
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
    )
    Barrier2 = og.create_load(
        loadtype="line", name="Barrieload", point1=barrierpoint_3, point2=barrierpoint_4
    )
    Barrier3 = og.create_load(
        loadtype="line", name="Barrieload", point1=barrierpoint_5, point2=barrierpoint_6
    )
    Patch1 = og.create_load(
        loadtype="patch",
        point1=barrierpoint_1,
        point2=barrierpoint_3,
        point3=barrierpoint_2,
        point4=barrierpoint_4,
    )

    barrierpoint_1 = og.create_load_vertex(x=6, z=2, p=0)
    barrierpoint_2 = og.create_load_vertex(x=11, z=8, p=0)
    barrierpoint_3 = og.create_load_vertex(x=11, z=3, p=0)
    barrierpoint_4 = og.create_load_vertex(x=6, z=6, p=0)

    Patch2 = og.create_load(
        loadtype="patch",
        point1=barrierpoint_1,
        point2=barrierpoint_3,
        point3=barrierpoint_2,
        point4=barrierpoint_4,
    )
    # create basic load case
    barrier_load_case = og.create_load_case(name="Barrier")
    # barrier_load_case.add_load(Barrier)  # ch
    barrier_load_case.add_load(Patch1)  # ch
    # 2nd
    barrier_load_case2 = og.create_load_case(name="Barrier2")
    # barrier_load_case2.add_load(Barrier2)
    barrier_load_case2.add_load(Patch2)
    # 3rd
    # barrier_load_case3 = og.create_load_case(name="Barrier3")
    # barrier_load_case3.add_load(Barrier3)

    # adding load cases to model
    example_bridge.add_load_case(barrier_load_case)
    example_bridge.add_load_case(barrier_load_case2)
    # example_bridge.add_load_case(barrier_load_case3)

    single_path = og.create_moving_path(
        start_point=og.Point(2, 0, 2), end_point=og.Point(4, 0, 3), increments=3
    )  # create path object
    move_point = og.create_moving_load(name="single_moving_point")
    move_point.set_path(single_path)
    move_point.add_load(front_wheel)
    example_bridge.add_load_case(move_point)

    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    # og.opsv.section_force_diagram_3d("Mz",{})

    f = og.plot_force(
        ospgrillage_obj=example_bridge,
        result_obj=results,
        component="Mz",
        member="interior_main_beam",
        loadcase="single_moving_point at global position [4.00,0.00,3.00]",
    )


def test_shell_plot_force(shell_link_bridge):
    # test functionality of plot_force on a shell_beam model type
    shell_link_model = shell_link_bridge
    # create and add load case comprise of single point load
    P = 20e3
    point_load_location = og.create_load_vertex(
        x=4.5, y=0, z=6.5, p=P
    )  # about midspan of span 1
    point_load = og.create_load(
        loadtype="point", name="single point", point1=point_load_location
    )
    point_lc = og.create_load_case(name="pointload")
    point_lc.add_load(point_load)
    shell_link_model.add_load_case(point_lc)
    shell_link_model.analyze()
    # extract results
    result = shell_link_model.get_results()
    print(result)
    f = og.plot_force(
        ospgrillage_obj=shell_link_model,
        result_obj=result,
        component="Mx",
        member="exterior_main_beam_1",
    )


def test_displacement_getter(bridge_model_42_negative):
    # test functionality of plot_force and its output
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative
    # og.opsv.plot_model()
    # og.plt.show()

    # add moving load case
    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )

    point_load_case = og.create_load_case(name="Point")
    point_load_case.add_load(front_wheel)
    example_bridge.add_load_case(point_load_case)

    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    processor = og.PostProcessor(grillage=example_bridge, result=results)

    arbitrary_disp = processor.get_arbitrary_displacements(point=[5, 0, 3])  # bigger

    assert all(np.isclose(arbitrary_disp * 1e6, [3.1289e-05 * 1e6]))


# ---------------------------------------------------------------------------
# og.plt accessibility
# ---------------------------------------------------------------------------
def test_og_plt_accessible():
    """og.plt should be matplotlib.pyplot."""
    assert hasattr(og, "plt")
    fig = og.plt.figure()
    assert fig is not None
    og.plt.close(fig)


# ---------------------------------------------------------------------------
# Convenience plotting wrappers
# ---------------------------------------------------------------------------
def test_plot_bmd(bridge_model_42_negative):
    """plot_bmd returns a figure for a single member and a list for all."""
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )
    lc = og.create_load_case(name="Point")
    lc.add_load(front_wheel)
    example_bridge.add_load_case(lc)
    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    # single member
    ax = og.plot_bmd(example_bridge, results, members="interior_main_beam")
    assert ax is not None

    # all main beams (returns list)
    axes = og.plot_bmd(example_bridge, results)
    assert isinstance(axes, list)
    assert len(axes) >= 1


def test_plot_sfd(bridge_model_42_negative):
    """plot_sfd returns a figure for a single member and a list for all."""
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )
    lc = og.create_load_case(name="Point")
    lc.add_load(front_wheel)
    example_bridge.add_load_case(lc)
    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    ax = og.plot_sfd(example_bridge, results, members="interior_main_beam")
    assert ax is not None

    axes = og.plot_sfd(example_bridge, results)
    assert isinstance(axes, list)
    assert len(axes) >= 1


def test_plot_def(bridge_model_42_negative):
    """plot_def returns a figure for a single member and a list for all."""
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )
    lc = og.create_load_case(name="Point")
    lc.add_load(front_wheel)
    example_bridge.add_load_case(lc)
    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    ax = og.plot_def(example_bridge, results, members="interior_main_beam")
    assert ax is not None

    axes = og.plot_def(example_bridge, results)
    assert isinstance(axes, list)
    assert len(axes) >= 1


# ---------------------------------------------------------------------------
# Plotly backend tests
# ---------------------------------------------------------------------------
def test_plot_bmd_plotly(bridge_model_42_negative):
    """plot_bmd with backend='plotly' returns a plotly Figure."""
    go = pytest.importorskip("plotly.graph_objects")
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )
    lc = og.create_load_case(name="Point")
    lc.add_load(front_wheel)
    example_bridge.add_load_case(lc)
    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    fig = og.plot_bmd(example_bridge, results, backend="plotly", show=False)
    assert isinstance(fig, go.Figure)
    assert len(fig.data) > 0  # has traces


def test_plot_sfd_plotly(bridge_model_42_negative):
    """plot_sfd with backend='plotly' returns a plotly Figure."""
    go = pytest.importorskip("plotly.graph_objects")
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )
    lc = og.create_load_case(name="Point")
    lc.add_load(front_wheel)
    example_bridge.add_load_case(lc)
    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    fig = og.plot_sfd(example_bridge, results, backend="plotly", show=False)
    assert isinstance(fig, go.Figure)
    assert len(fig.data) > 0


def test_plot_def_plotly(bridge_model_42_negative):
    """plot_def with backend='plotly' returns a plotly Figure."""
    go = pytest.importorskip("plotly.graph_objects")
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )
    lc = og.create_load_case(name="Point")
    lc.add_load(front_wheel)
    example_bridge.add_load_case(lc)
    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    fig = og.plot_def(example_bridge, results, backend="plotly", show=False)
    assert isinstance(fig, go.Figure)
    assert len(fig.data) > 0


def test_plotly_single_member(bridge_model_42_negative):
    """plot_bmd with plotly and a single member should return a Figure."""
    go = pytest.importorskip("plotly.graph_objects")
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )
    lc = og.create_load_case(name="Point")
    lc.add_load(front_wheel)
    example_bridge.add_load_case(lc)
    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    fig = og.plot_bmd(
        example_bridge,
        results,
        members="interior_main_beam",
        backend="plotly",
        show=False,
    )
    assert isinstance(fig, go.Figure)


def test_plotly_import_error():
    """When plotly is unavailable, backend='plotly' should raise ImportError."""
    from unittest.mock import patch
    from ospgrillage.postprocessing import _import_plotly

    with patch.dict("sys.modules", {"plotly": None, "plotly.graph_objects": None}):
        with pytest.raises(ImportError, match="ospgrillage\\[gui\\]"):
            _import_plotly()


# ---------------------------------------------------------------------------
# Plotting kwargs tests
# ---------------------------------------------------------------------------
def _make_analyzed_bridge(bridge_model_42_negative):
    """Helper: set up a load case, analyse, and return (bridge, results)."""
    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )
    lc = og.create_load_case(name="Point")
    lc.add_load(front_wheel)
    bridge.add_load_case(lc)
    bridge.analyze()
    results = bridge.get_results(local_forces=False)
    return bridge, results


def test_plot_force_figsize(bridge_model_42_negative):
    """figsize is forwarded to plt.subplots."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    ax = og.plot_force(
        bridge,
        results,
        component="Mz",
        member="interior_main_beam",
        figsize=(10, 4),
    )
    w, h = ax.get_figure().get_size_inches()
    assert abs(w - 10) < 0.1
    assert abs(h - 4) < 0.1


def test_plot_force_ax(bridge_model_42_negative):
    """Passing an existing Axes should reuse the axes."""
    import matplotlib.pyplot as plt

    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)

    fig_ext, ax_ext = plt.subplots(figsize=(12, 3))
    returned_ax = og.plot_force(
        bridge,
        results,
        component="Mz",
        member="interior_main_beam",
        ax=ax_ext,
    )
    assert returned_ax is ax_ext


def test_plot_force_scale(bridge_model_42_negative):
    """scale=2.0 should double the plotted y-values."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)

    ax1 = og.plot_force(
        bridge,
        results,
        component="Mz",
        member="interior_main_beam",
        scale=1.0,
    )
    ax2 = og.plot_force(
        bridge,
        results,
        component="Mz",
        member="interior_main_beam",
        scale=2.0,
    )
    # Compare the first line's y-data
    import numpy as np

    y1 = ax1.lines[0].get_ydata()
    y2 = ax2.lines[0].get_ydata()
    np.testing.assert_allclose(y2, y1 * 2, rtol=1e-10)


def test_plot_force_styling(bridge_model_42_negative):
    """Smoke test: color, fill, alpha, title, show kwargs don't error."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    ax = og.plot_force(
        bridge,
        results,
        component="Mz",
        member="interior_main_beam",
        color="r",
        fill=False,
        alpha=0.8,
        title="Custom Title",
        show=False,
    )
    assert ax.get_title() == "Custom Title"
    # fill=False means no PolyCollection (only lines)
    assert len(ax.collections) == 0


def test_plot_force_title_none(bridge_model_42_negative):
    """title=None should suppress the title."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    ax = og.plot_force(
        bridge,
        results,
        component="Mz",
        member="interior_main_beam",
        title=None,
    )
    assert ax.get_title() == ""


def test_plot_bmd_kwargs_passthrough(bridge_model_42_negative):
    """plot_bmd forwards kwargs to plot_force without error."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    ax = og.plot_bmd(
        bridge,
        results,
        members="interior_main_beam",
        figsize=(12, 4),
        scale=0.001,
        title="BMD (kNm)",
    )
    assert ax is not None
    assert ax.get_title() == "BMD (kNm)"


def test_plot_def_kwargs_passthrough(bridge_model_42_negative):
    """plot_def forwards kwargs to the deflection renderer without error."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    ax = og.plot_def(
        bridge,
        results,
        members="interior_main_beam",
        figsize=(12, 4),
        scale=1000,
        color="g",
    )
    assert ax is not None


def test_plotly_kwargs(bridge_model_42_negative):
    """Plotly backend accepts figsize, scale, title kwargs."""
    go = pytest.importorskip("plotly.graph_objects")
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)

    fig = og.plot_bmd(
        bridge,
        results,
        backend="plotly",
        show=False,
        figsize=(12, 8),
        scale=0.5,
        title="Custom Plotly BMD",
    )
    assert isinstance(fig, go.Figure)
    assert fig.layout.title.text == "Custom Plotly BMD"
    assert fig.layout.width == 1200
    assert fig.layout.height == 800


# ---------------------------------------------------------------------------
# plot_model tests
# ---------------------------------------------------------------------------
def test_plot_model_matplotlib(bridge_model_42_negative):
    """plot_model with matplotlib returns an Axes showing the mesh."""
    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    ax = og.plot_model(bridge)
    assert ax is not None
    assert len(ax.lines) > 0


def test_plot_model_matplotlib_labels(bridge_model_42_negative):
    """plot_model shows node/element labels when requested."""
    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    ax = og.plot_model(bridge, show_node_labels=True, show_element_labels=True)
    assert len(ax.texts) > 0


def test_plot_model_matplotlib_kwargs(bridge_model_42_negative):
    """plot_model accepts figsize, title, ax kwargs."""
    import matplotlib.pyplot as plt

    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    ax = og.plot_model(bridge, figsize=(12, 6), title="Test Model")
    w, h = ax.get_figure().get_size_inches()
    assert abs(w - 12) < 0.1
    assert ax.get_title() == "Test Model"


def test_plot_model_plotly(bridge_model_42_negative):
    """plot_model with plotly returns a plotly Figure."""
    go = pytest.importorskip("plotly.graph_objects")
    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    fig = og.plot_model(bridge, backend="plotly", show=False)
    assert isinstance(fig, go.Figure)
    assert len(fig.data) > 0


def test_plot_model_invalid_backend(bridge_model_42_negative):
    """plot_model raises ValueError for unknown backend."""
    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    with pytest.raises(ValueError, match="Unknown backend"):
        og.plot_model(bridge, backend="vtk")


# ---------------------------------------------------------------------------
# Members enum tests
# ---------------------------------------------------------------------------
def test_members_enum_composites():
    """Members composites resolve to expected individual flags."""
    assert og.Members.LONGITUDINAL == (
        og.Members.EDGE_BEAM
        | og.Members.EXTERIOR_MAIN_BEAM_1
        | og.Members.INTERIOR_MAIN_BEAM
        | og.Members.EXTERIOR_MAIN_BEAM_2
    )
    assert og.Members.TRANSVERSE == (
        og.Members.START_EDGE | og.Members.END_EDGE | og.Members.TRANSVERSE_SLAB
    )
    assert og.Members.ALL == og.Members.LONGITUDINAL | og.Members.TRANSVERSE


def test_resolve_members_none_plotly():
    """member=None with plotly backend resolves to all 7 members."""
    from ospgrillage.postprocessing import _resolve_members

    result = _resolve_members(None, backend="plotly")
    assert len(result) == 7
    assert "transverse_slab" in result
    assert "start_edge" in result


def test_resolve_members_none_matplotlib():
    """member=None with matplotlib backend resolves to 4 longitudinal members."""
    from ospgrillage.postprocessing import _resolve_members

    result = _resolve_members(None, backend="matplotlib")
    assert len(result) == 4
    assert "transverse_slab" not in result


def test_resolve_members_string():
    """A string member name resolves to a single-element list."""
    from ospgrillage.postprocessing import _resolve_members

    result = _resolve_members("interior_main_beam")
    assert result == ["interior_main_beam"]


def test_resolve_members_flag_combination():
    """A combined flag resolves in declaration order."""
    from ospgrillage.postprocessing import _resolve_members

    result = _resolve_members(og.Members.TRANSVERSE_SLAB | og.Members.EDGE_BEAM)
    assert result == ["edge_beam", "transverse_slab"]


def test_resolve_members_invalid_string():
    """An unknown member name raises ValueError."""
    from ospgrillage.postprocessing import _resolve_members

    with pytest.raises(ValueError, match="Unknown member"):
        _resolve_members("nonexistent_beam")


def test_resolve_members_invalid_type():
    """An unsupported type raises TypeError."""
    from ospgrillage.postprocessing import _resolve_members

    with pytest.raises(TypeError, match="must be str, Members, or None"):
        _resolve_members(42)


def test_plot_bmd_with_members_flag(bridge_model_42_negative):
    """plot_bmd matplotlib backend accepts a Members flag."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    axes = og.plot_bmd(
        bridge,
        results,
        members=og.Members.INTERIOR_MAIN_BEAM | og.Members.EDGE_BEAM,
    )
    assert isinstance(axes, list)
    assert len(axes) >= 1


def test_plot_sfd_with_members_flag(bridge_model_42_negative):
    """plot_sfd matplotlib backend accepts a Members flag."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    axes = og.plot_sfd(
        bridge,
        results,
        members=og.Members.LONGITUDINAL,
    )
    assert isinstance(axes, list)
    assert len(axes) >= 1


def test_plot_def_with_members_flag(bridge_model_42_negative):
    """plot_def matplotlib backend accepts a Members flag."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    axes = og.plot_def(
        bridge,
        results,
        members=og.Members.INTERIOR_MAIN_BEAM,
    )
    assert isinstance(axes, list)
    assert len(axes) >= 1


def test_plot_bmd_plotly_members_flag(bridge_model_42_negative):
    """plot_bmd plotly backend accepts a Members flag."""
    go = pytest.importorskip("plotly.graph_objects")
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    fig = og.plot_bmd(
        bridge,
        results,
        members=og.Members.LONGITUDINAL,
        backend="plotly",
        show=False,
    )
    assert isinstance(fig, go.Figure)
    assert len(fig.data) > 0


def test_plot_def_plotly_members_flag(bridge_model_42_negative):
    """plot_def plotly backend accepts a Members flag."""
    go = pytest.importorskip("plotly.graph_objects")
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)
    fig = og.plot_def(
        bridge,
        results,
        members=og.Members.ALL,
        backend="plotly",
        show=False,
    )
    assert isinstance(fig, go.Figure)
    assert len(fig.data) > 0


# ---------------------------------------------------------------------------
# Torsion moment diagram tests
# ---------------------------------------------------------------------------
def test_plot_tmd(bridge_model_42_negative):
    """plot_tmd returns a figure for a single member and a list for all."""
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)

    # single member
    ax = og.plot_tmd(bridge, results, members="interior_main_beam")
    assert ax is not None

    # all main beams (returns list)
    axes = og.plot_tmd(bridge, results)
    assert isinstance(axes, list)
    assert len(axes) >= 1


def test_plot_tmd_plotly(bridge_model_42_negative):
    """plot_tmd with backend='plotly' returns a plotly Figure."""
    go = pytest.importorskip("plotly.graph_objects")
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)

    fig = og.plot_tmd(bridge, results, backend="plotly", show=False)
    assert isinstance(fig, go.Figure)
    assert len(fig.data) > 0


# ---------------------------------------------------------------------------
# Plotly fill (Mesh3d ribbon) tests
# ---------------------------------------------------------------------------
def test_plotly_force_fill(bridge_model_42_negative):
    """Plotly BMD includes Mesh3d fill traces by default."""
    go = pytest.importorskip("plotly.graph_objects")
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)

    fig = og.plot_bmd(
        bridge,
        results,
        members="interior_main_beam",
        backend="plotly",
        show=False,
    )
    mesh_traces = [t for t in fig.data if isinstance(t, go.Mesh3d)]
    assert len(mesh_traces) >= 1


def test_plotly_force_no_fill(bridge_model_42_negative):
    """fill=False suppresses Mesh3d traces."""
    go = pytest.importorskip("plotly.graph_objects")
    bridge, results = _make_analyzed_bridge(bridge_model_42_negative)

    fig = og.plot_bmd(
        bridge,
        results,
        members="interior_main_beam",
        backend="plotly",
        show=False,
        fill=False,
    )
    mesh_traces = [t for t in fig.data if isinstance(t, go.Mesh3d)]
    assert len(mesh_traces) == 0


# ---------------------------------------------------------------------------
# Support drawing tests
# ---------------------------------------------------------------------------
def test_plot_model_supports_matplotlib(bridge_model_42_negative):
    """plot_model with show_supports=True draws support markers."""
    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    ax = og.plot_model(bridge, show_supports=True)
    assert ax is not None


def test_plot_model_supports_plotly(bridge_model_42_negative):
    """plot_model plotly with show_supports=True includes support traces."""
    go = pytest.importorskip("plotly.graph_objects")
    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    fig = og.plot_model(bridge, backend="plotly", show=False, show_supports=True)
    assert isinstance(fig, go.Figure)
    support_traces = [t for t in fig.data if "support" in (t.name or "")]
    assert len(support_traces) >= 1


def test_plot_model_no_supports(bridge_model_42_negative):
    """plot_model with show_supports=False omits support markers."""
    go = pytest.importorskip("plotly.graph_objects")
    og.ops.wipeAnalysis()
    bridge = bridge_model_42_negative
    fig = og.plot_model(bridge, backend="plotly", show=False, show_supports=False)
    support_traces = [t for t in fig.data if "support" in (t.name or "")]
    assert len(support_traces) == 0



# -*- coding: utf-8 -*-
"""
This module contains the user interface function and class to manage
 material properties that allow definition offine grillage elements.
 The module also contains methods that wraps ```OpenSeesPy``` material object creation.
"""

import json
import logging

logger = logging.getLogger(__name__)

__all__ = ["Material", "create_material"]


def create_material(**kwargs):
    """
    Create a material object for use in a grillage model.

    The constructor of :class:`Material` object takes in three types of keyword arguments:

    #. Keyword for looking up the *ospgrillage* material library i.e. mat_lib.json.
    #. General material properties - such as E, and G
    #. Specific arguments of ``OpenSeesPy`` material library.


    For material library lookup, provide:

    :param code: Code name string as defined in ``mat_lib.json`` (e.g. ``"AS5100"``).
    :type code: str, optional
    :param type: Material type — either ``"concrete"`` or ``"steel"``.
    :type type: str, optional
    :param grade: Material grade according to the selected code.
    :type grade: str, optional

    For general material properties, provide:

    :param E: Elastic modulus.
    :type E: float, optional
    :param G: Shear modulus.
    :type G: float, optional

    """
    return Material(**kwargs)


class Material:
    """
    This class stores information and provides methods to parse input material properties into ``OpenSeesPy`` Material
    objects.

    `Here <https://openseespydoc.readthedocs.io/en/latest/src/uniaxialMaterial.html>`_ are the information about OpenSees
    Material objects.

    As *ospgrillage* is mainly intended for bridge decks , concrete and steel makes up the primary
    materials. In turn, *ospgrillage* wraps the UniAxialMaterial object of ``OpenSeesPy`` since it contains the primary
    options for Concrete and Steel.

    The Material class also allow users to utilize codified material properties (e.g. AS5100, ASHTOO).
    These material properties are stored in a material library file (mat_lib.json). Users are required to pass in
    the appropriate keyword arguments (dict keys) to select the desirable codified materials. Users may access
    mat_lib.json. Additionally, this class is able create the mat_lib.json file and pass it through

    .. note::

        Current version of mat_lib.json is 0.1.0
    """

    def __init__(self, **kwargs):
        """
        The constructor of Material takes in three types of inputs:

        #. Keywords for looking up the *ospgrillage* material library i.e. mat_lib.json.
        #. General material properties - such as E, and G
        #. Specific material arguments of ``OpenSeesPy``.

        For material library lookup (option 1), provide:

        :param code: Code name string as defined in ``mat_lib.json``.
        :type code: str, optional
        :param type: Material type — either ``"concrete"`` or ``"steel"``.
        :type type: str, optional
        :param grade: Material grade according to the selected code.
        :type grade: str, optional

        For general material properties (option 2), provide:

        :param E: Elastic modulus.
        :type E: float, optional
        :param G: Shear modulus.
        :type G: float, optional
        :param v: Poisson's ratio.
        :type v: float, optional
        :param rho: Density.
        :type rho: float, optional

        For OpenSeesPy-specific arguments (option 3), refer to
        `OpenSeesPy <https://openseespydoc.readthedocs.io/en/latest/src/uniaxialMaterial.html>`_.


        """
        # Instantiate variables
        self.ops_mat_type = kwargs.get(
            "ops_mat_type", None
        )  # this is the os convention for materials e.g. Concrete01
        self.op_mat_arg = None  # arguments according to OpenSeesPy
        self.units = kwargs.get("units", None)  # default SI units
        # assigns variables for all kwargs for codified material types
        self.code = kwargs.get("code", None)
        self.material_type = kwargs.get("material", None)
        self.material_grade = kwargs.get("grade", None)
        # assign generic material properties
        self.elastic_modulus = kwargs.get("E", None)
        self.shear_modulus = kwargs.get("G", None)
        self.poisson_ratio = kwargs.get("v", None)
        self.density = kwargs.get("rho", None)

        # get mat lib file
        self.default_mat = kwargs.get("default_material", True)
        self._mat_lib = self._read_mat_lib()
        # ----------------------------------------------------------------------------------
        # material vars

        # properties for Standard Elastic OpenSeesPy material

        # properties for Concrete - symbols according to OpenSees uniaxialMaterial/Concrete
        self.fpc = kwargs.get("fpc", None)
        self.epsc0 = kwargs.get("epsc0", None)
        self.fpcu = kwargs.get("fpcu", None)
        self.epsU = kwargs.get("epsU", None)

        # properties for Steel - symbols according to OpenSees uniaxialMaterial/Steel
        self.Fy = kwargs.get("Fy", None)
        self.E0 = kwargs.get("E0", None)
        self.b = kwargs.get("b", None)  # strain -hardening ratio.
        self.a1 = kwargs.get("a1", None)
        self.a2 = kwargs.get(
            "a2", None
        )  # isotropic hardening parameter , see Ops Steel01
        self.a3 = kwargs.get("a3", None)
        self.a4 = kwargs.get("a4", None)

        # ----------------------------------------------------------------------------------
        # process all inputs into relevant inputs for OpenSees material command
        self._parse_material_command()

    def _parse_material_command(self):
        """
        Parse the material properties inputs
        """
        # check if code material is selected, if yes read from material library json
        if self.code:
            self.poisson_ratio = self._mat_lib[self.material_type][self.code][
                self.material_grade
            ]["v"]
            self.elastic_modulus = self._mat_lib[self.material_type][self.code][
                self.material_grade
            ]["E"]
            self.fpc = self._mat_lib[self.material_type][self.code][
                self.material_grade
            ]["fc"]
            self.density = self._mat_lib[self.material_type][self.code][
                self.material_grade
            ]["rho"]
            self.units = self._mat_lib[self.material_type][self.code]["units"]

        # perform conversion for units
        if self.units == "SI":
            self.elastic_modulus = self.elastic_modulus * 1e9
            self.fpc = self.fpc * 1e6

        # check poison ratio
        if not self.poisson_ratio:
            self.poisson_ratio = 0.3

        # if G not defined, calculate using formula E/(2(1+v))
        if self.shear_modulus is None:
            self.shear_modulus = self.elastic_modulus / (2 * (1 + self.poisson_ratio))

        if self.material_type == "concrete":
            self.ops_mat_type = (
                "Concrete01"  # default opensees material type to represent concrete
            )
        elif self.material_type == "steel":
            self.ops_mat_type = (
                "Steel01"  # default opensees material type to represent steel
            )

    def get_material_args(self):
        """
        Return the arguments for OpenSeesPy material. This function is handled by
        :class:`ospgrillage.osp_grillage.OspGrillage`.

        :returns: Str of OpenSeesPy material type and list of material properties correspond to the inputs of
         OpenSeesPy commands
        """
        if self.ops_mat_type == "Concrete01":
            self.op_mat_arg = [self.fpc, self.epsc0, self.fpcu, self.epsU]
        elif self.ops_mat_type == "Steel01":
            self.op_mat_arg = [
                self.Fy,
                self.E0,
                self.b,
                self.a1,
                self.a2,
                self.a3,
                self.a4,
            ]
        elif self.ops_mat_type == "Elastic":
            self.op_mat_arg = [self.elastic_modulus]
        # TO ADD for MORE materials

        # check if None in entries
        if None in self.op_mat_arg:
            raise ValueError(
                "One or more missing/non-numeric parameters for Material: {} ".format(
                    self.ops_mat_type
                )
            )
        return self.ops_mat_type, self.op_mat_arg

    @staticmethod
    def _create_default_dict():
        """
        Create the default mat_lib.js file. The default version is 0.1.0.
        Just to make sure the JSON file is formatted correctly
        Note: 1 ksi = 6.89475728 MPa
        """

        mat_lib = {
            "concrete": {
                "AS5100-2017": {
                    "units": "SI",
                    "25MPa": {"fc": 25, "E": 26.7, "v": 0.2, "rho": 2.4e3},
                    "32MPa": {"fc": 32, "E": 30.1, "v": 0.2, "rho": 2.4e3},
                    "40MPa": {"fc": 40, "E": 32.8, "v": 0.2, "rho": 2.4e3},
                    "50MPa": {"fc": 50, "E": 34.8, "v": 0.2, "rho": 2.4e3},
                    "65MPa": {"fc": 65, "E": 37.4, "v": 0.2, "rho": 2.4e3},
                    "80MPa": {"fc": 80, "E": 39.6, "v": 0.2, "rho": 2.4e3},
                    "100MPa": {"fc": 100, "E": 42.2, "v": 0.2, "rho": 2.4e3},
                },
                "AASHTO-LRFD-8th": {
                    "units": "SI",
                    "2.4ksi": {"fc": 16.55, "E": 23.2223, "v": 0.2, "rho": 2.4027e3},
                    "3.0ksi": {"fc": 20.68, "E": 24.997, "v": 0.2, "rho": 2.4027e3},
                    "3.6ksi": {"fc": 24.82, "E": 26.547, "v": 0.2, "rho": 2.4027e3},
                    "4.0ksi": {"fc": 27.58, "E": 27.486, "v": 0.2, "rho": 2.4027e3},
                    "5.0ksi": {"fc": 34.47, "E": 29.587, "v": 0.2, "rho": 2.4027e3},
                    "6.0ksi": {"fc": 41.37, "E": 31.856, "v": 0.2, "rho": 2.4027e3},
                    "7.5ksi": {"fc": 51.71, "E": 34.999, "v": 0.2, "rho": 2.4027e3},
                    "10.0ksi": {"fc": 68.95, "E": 39.8, "v": 0.2, "rho": 2.4027e3},
                    "15.0ksi": {"fc": 103.42, "E": 48.582, "v": 0.2, "rho": 2.4027e3},
                },
            },
            "steel": {
                "AS5100.6-2004": {
                    "units": "SI",
                    "R250N": {"fy": 250, "E": 200.0, "v": 0.25, "rho": 7850},
                    "D500N": {"fy": 500, "E": 200.0, "v": 0.25, "rho": 7850},
                    "D500L": {"fy": 500, "E": 200.0, "v": 0.25, "rho": 7850},
                },
                "AASHTO-LRFD-8th": {
                    "units": "SI",
                    "A615-40": {"fy": 275.8, "E": 200.0, "v": 0.3, "rho": 7849},
                    "A615-60": {"fy": 413.67, "E": 200.0, "v": 0.3, "rho": 7849},
                    "A615-75": {"fy": 517.12, "E": 200.0, "v": 0.3, "rho": 7849},
                    "A615-80": {"fy": 551.58, "E": 200.0, "v": 0.3, "rho": 7849},
                    "A615-100": {"fy": 689.48, "E": 200.0, "v": 0.3, "rho": 7849},
                },
            },
        }

        return mat_lib

    @staticmethod
    def _write_mat_lib(mat_lib):
        """
        Write out the passed material dict
        Not to be used in the ordinary course of events at risk of overwriting
        manual edits to the mat_lib
        Used for initial creation of the mat_lib
        """

        with open("mat_lib.json", "w") as f:
            json.dump(mat_lib, f, indent=4)

    def _read_mat_lib(self):
        """
        Read material library from json file
        """
        mat_lib = {}
        try:
            with open("mat_lib.json", "r") as f:
                mat_lib = json.load(f)
        except (FileNotFoundError, IOError):
            logger.warning("Material library unable to be read; using default library")
            mat_lib = self._create_default_dict()
            self._write_mat_lib(mat_lib)
        return mat_lib

    def get_ops_material_command(self, material_tag):
        """
        Parse material properties into OpenSeesPy Material commands.

        :param material_tag: tag of material defined in in OpenSeesPy space
        :type material_tag: int

        :returns: Str of OpenSees command to create material in OpenSeesPy

        """
        # e.g. concrete01 or steel01
        mat_str = None
        if (
            self.ops_mat_type == "Concrete01"
            or self.ops_mat_type == "Steel01"
            or self.ops_mat_type == "Elastic"
        ):
            mat_str = 'ops.uniaxialMaterial("{type}", {tag}, *{vec})\n'.format(
                type=self.ops_mat_type, tag=material_tag, vec=self.op_mat_arg
            )

        return mat_str
